from math import ceil

import numpy as np

np.seterr(all="raise")


def return_one_percent(num, pop_size):
    """
    Returns either one percent of the population size or a given number,
    whichever is larger.

    :param num: A given number of individuals (NOT a desired percentage of
    the population).
    :param pop_size: A given population size.
    :return: either one percent of the population size or a given number,
    whichever is larger.
    """

    # Calculate one percent of the given population size.
    percent = int(round(pop_size / 100))

    # Return the biggest number.
    if percent < num:
        return num
    else:
        return percent


def return_percent(num, pop_size):
    """
    Returns [num] percent of the population size.

    :param num: A desired percentage of the population.
    :param pop_size: A given population size.
    :return: [num] percent of the population size.
    """

    return int(round(num * pop_size / 100))


def aq(a, b):
    """aq is the analytic quotient, intended as a "better protected
    division", from: Ji Ni and Russ H. Drieberg and Peter I. Rockett,
    "The Use of an Analytic Quotient Operator in Genetic Programming",
    IEEE Transactions on Evolutionary Computation.

    :param a: np.array numerator
    :param b: np.array denominator
    :return: np.array analytic quotient, analogous to a / b.

    """
    return a / np.sqrt(1.0 + b ** 2.0)


def pdiv(x, y):
    """
    Koza's protected division is:

    if y == 0:
      return 1
    else:
      return x / y

    but we want an eval-able expression. The following is eval-able:

    return 1 if y == 0 else x / y

    but if x and y are Numpy arrays, this creates a new Boolean
    array with value (y == 0). if doesn't work on a Boolean array.

    The equivalent for Numpy is a where statement, as below. However
    this always evaluates x / y before running np.where, so that
    will raise a 'divide' error (in Numpy's terminology), which we
    ignore using a context manager.

    In some instances, Numpy can raise a FloatingPointError. These are
    ignored with 'invalid = ignore'.

    :param x: numerator np.array
    :param y: denominator np.array
    :return: np.array of x / y, or 1 where y is 0.
    """
    try:
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.where(y == 0, np.ones_like(x), x / y)
    except ZeroDivisionError:
        # In this case we are trying to divide two constants, one of which is 0
        # Return a constant.
        return 1.0


def rlog(x):
    """
    Koza's protected log:
    if x == 0:
      return 1
    else:
      return log(abs(x))

    See pdiv above for explanation of this type of code.

    :param x: argument to log, np.array
    :return: np.array of log(x), or 1 where x is 0.
    """
    with np.errstate(divide="ignore"):
        return np.where(x == 0, np.ones_like(x), np.log(np.abs(x)))


def ppow(x, y):
    """pow(x, y) is undefined in the case where x negative and y
    non-integer. This takes abs(x) to avoid it.

    :param x: np.array, base
    :param y: np.array, exponent
    :return: np.array x**y, but protected

    """
    return np.abs(x) ** y


def ppow2(x, y):
    """pow(x, y) is undefined in the case where x negative and y
    non-integer. This takes abs(x) to avoid it. But it preserves
    sign using sign(x).

    :param x: np.array, base
    :param y: np.array, exponent
    :return: np.array, x**y, but protected
    """
    return np.sign(x) * (np.abs(x) ** y)


def psqrt(x):
    """
    Protected square root operator

    :param x: np.array, argument to sqrt
    :return: np.array, sqrt(x) but protected.
    """
    return np.sqrt(np.abs(x))


def psqrt2(x):
    """
    Protected square root operator that preserves the sign of the original
    argument.

    :param x: np.array, argument to sqrt
    :return: np.array, sqrt(x) but protected, preserving sign.
    """
    return np.sign(x) * (np.sqrt(np.abs(x)))


def plog(x):
    """
    Protected log operator. Protects against the log of 0.

    :param x: np.array, argument to log
    :return: np.array of log(x), but protected
    """
    return np.log(1.0 + np.abs(x))


def ave(x):
    """
    Returns the average value of a list.

    :param x: a given list
    :return: the average of param x
    """

    return np.mean(x)


def percentile(sorted_list, p):
    """
    Returns the element corresponding to the p-th percentile
    in a sorted list

    :param sorted_list: The sorted list
    :param p: The percentile
    :return: The element corresponding to the percentile
    """

    return sorted_list[ceil(len(sorted_list) * p / 100) - 1]


def binary_phen_to_float(phen, n_codon, min_value, max_value):
    """
    This method converts a phenotype, defined by a
    string of bits in a list of float values

    :param phen: Phenotype defined by a bit string
    :param n_codon: Number of codons per gene, defined in the grammar
    :param min_value: Minimum value for a gene
    :param max_value: Maximum value for a gene
    :return: A list os float values, representing the chromosome
    """

    i, count, chromosome = 0, 0, []

    while i < len(phen):
        # Get the current gene from the phenotype string.
        gene = phen[i: (i + n_codon)]

        # Convert the bit string in gene to an float/int
        gene_i = int(gene, 2)
        gene_f = float(gene_i) / (2 ** n_codon - 1)

        # Define the variation for the gene
        delta = max_value[count] - min_value[count]

        # Append the float value to the chromosome list
        chromosome.append(gene_f * delta + min_value[count])

        # Increment the index and count.
        i = i + n_codon
        count += 1

    return chromosome


def ilog(n, base):
    """
    Find the integer log of n with respect to the base.

    >>> import math
    >>> for base in range(2, 16 + 1):
    ...     for n in range(1, 1000):
    ...         assert ilog(n, base) == int(math.log(n, base) + 1e-10), '%s %s' % (n, base)
    """
    count = 0
    while n >= base:
        count += 1
        n //= base
    return count


def sci_notation(n, prec=3):
    """
    Represent n in scientific notation, with the specified precision.

    >>> sci_notation(1234 * 10**1000)
    '1.234e+1003'
    >>> sci_notation(10**1000 // 2, prec=1)
    '5.0e+999'
    """
    base = 10
    exponent = ilog(n, base)
    mantissa = n / base ** exponent
    return "{0:.{1}f}e{2:+d}".format(mantissa, prec, exponent)



import itertools
import random

import numpy as np

from autooc.algorithm.parameters import params
from autooc.fitness.supervised_learning.supervised_learning import \
    supervised_learning
from autooc.utilities.fitness.error_metric import Hamming_error


class boolean_problem(supervised_learning):
    """Fitness function for Boolean problems. We specialise the
    supervised learning fitness function.

    The user must pass in the name of the target function, and n_vars.
    The function is created. It is evaluated at all possible
    inputs. That gives a target dataset for training. This is
    accomplished with, eg, --extra_parameters nparity 5. There is
    no test on unseen data.

    An example command-line is then:

    python ponyge.py --generations 10 --population 10 --fitness
    supervised_learning.boolean_problem
    --extra_parameters nparity 3 --grammar
    supervised_learning/boolean.bnf

    """

    def __init__(self):
        # Don't call super().__init__() because it reads the training
        # (and test) data from files. We'll do everything else it
        # would have done.

        target_name = params["EXTRA_PARAMETERS"][0]

        # may be needed if the grammar uses GE_RANGE:dataset_n_vars
        n = self.n_vars = int(params["EXTRA_PARAMETERS"][1])

        # Set error metric if it's not set already.
        if params["ERROR_METRIC"] is None:
            params["ERROR_METRIC"] = Hamming_error

        self.maximise = params["ERROR_METRIC"].maximise

        # create the target function
        if target_name == "random_boolean":
            target = make_random_boolean_fn(n)
        else:
            target = eval(target_name)

        # generate all input combinations for n variables, to become
        # the fitness cases
        Ls = [[False, True] for i in range(n)]
        X = np.array(list(itertools.product(*Ls)))
        # evaluate the target function at the fitness cases
        self.training_exp = np.array([target(xi) for xi in X])
        self.training_in = X.T

        # In Boolean problems we don't want a separate test set
        assert not params["DATASET_TEST"]


# Some target functions. Each just accepts a single instance, eg
# nparity([False, False, True]) -> True


def boolean_true(x):
    return True


def comparator(x):
    """Comparator function: input consists of two n-bit numbers. Output is
    0 if the first is larger or equal, or 1 if the second is larger."""
    assert len(x) % 2 == 0
    n = len(x) // 2
    # no need to convert from binary. just use list comparison
    return x[:n] < x[n:]


def multiplexer(x):
    """Multiplexer: a address bits and 2^a data bits. n = a + 2^a. Output
    the value of the data bit addressed by the address bits.
    """
    if len(x) == 1 + 2 ** 1:
        a = 1  # 2
    elif len(x) == 2 + 2 ** 2:
        a = 2  # 6
    elif len(x) == 3 + 2 ** 3:
        a = 3  # 11
    elif len(x) == 4 + 2 ** 4:
        a = 4  # 20
    elif len(x) == 5 + 2 ** 5:
        a = 5  # 37
    elif len(x) == 6 + 2 ** 6:
        a = 6  # 70
    else:
        raise ValueError(x)
    addr = binlist2int(x[:a])  # get address bits, convert to int
    return x[a + addr]  # which data bit? offset by a


def nparity(x):
    "Parity function of any number of input variables"
    return x.sum() % 2 == 1


# special target function: random truth table
def make_random_boolean_fn(n):
    """Make a random Boolean function of n variables."""
    outputs = [random.choice([False, True]) for i in range(2 ** n)]

    def f(x):
        return outputs[binlist2int(x)]

    return f


# Helper function


def binlist2int(x):
    """Convert a list of binary digits to integer"""
    return int("".join(map(str, map(int, x))), 2)



import sys
from sys import path

from autooc.algorithm.parameters import params, set_params
from autooc.operators.subtree_parse import (check_snippets_for_solution,
                                           generate_key_and_check,
                                           get_NT_from_str, get_num_from_str)
from autooc.representation.tree import Tree
from autooc.utilities.algorithm.general import check_python_version
from autooc.utilities.representation.check_methods import (
    check_ind_from_parser, generate_codon)
from autooc.utilities.stats import trackers

path.append("../src")


check_python_version()


def parse_terminals(target):
    """
    Given a target string, build up a list of terminals which match certain
    portions of the target string.

    :return: A list of terminals in order of appearance in the target string.
    """

    if params["VERBOSE"]:
        print("Target:\n", target)

    # Pre-load all terminals and non-terminal rules from the grammar.
    terms, rules = params["BNF_GRAMMAR"].terminals, params["BNF_GRAMMAR"].rules

    # Initialise dict for storing the snippets for compiling a complete
    # solution. The key for each entry is the portion of the target string on
    # which the output matches, along with the root node of the subtree. The
    # value is the subtree itself.
    trackers.snippets = {}

    # Initialise dict for deleted snippets, to ensure they aren't generated
    # again.
    trackers.deleted_snippets = []

    for T in sorted(terms.keys()):
        # Iterate over all Terminals.

        # Find all occurrences of this terminal in the target string.
        occurrences = []
        index = 0
        while index < len(target):
            index = target.find(T, index)
            if index not in occurrences and index != -1:
                occurrences.append(index)
                index += len(T)
            else:
                break

        for idx in occurrences:
            # Check each occurrence of this terminal in the target string.

            for NT in terms[T]:

                if any(
                    [
                        [T] == i
                        for i in [
                            [sym["symbol"] for sym in choice["choice"]]
                            for choice in rules[NT]["choices"]
                        ]
                    ]
                ):
                    # Check if current T is the entire production choice of any
                    # particular rule.

                    # Generate a key for the snippets repository.
                    key = " ".join([str([idx, idx + len(T)]), NT])

                    # Get index of production choice.
                    index = [
                        [sym["symbol"] for sym in choice["choice"]]
                        for choice in rules[NT]["choices"]
                    ].index([T])

                    # Get production choice.
                    choice = rules[NT]["choices"][index]["choice"]

                    # Generate a tree for this choice.
                    parent = Tree(NT, None)

                    # Generate a codon for this choice.
                    parent.codon = generate_codon(NT, choice)

                    # Set the snippet key for the parent.
                    parent.snippet = key

                    # Create child for terminal.
                    child = Tree(T, parent)

                    # Add child to parent.
                    parent.children.append(child)

                    # Add snippet to snippets repository.
                    trackers.snippets[key] = parent


def reduce(solution):
    """
    Takes a list of all matching subtrees found in the target string and
    iteratively combines and reduces subtrees to generate larger matching
    subtrees. This process continues until the list of matching subtrees has
    been completely reduced into a target string.

    :param solution: A list of all snippets (i.e. matching subtrees found in
    the target string.
    :return: Nothing.
    """

    # Find all non-terminals in the grammar that can be used to concatenate
    # subtrees to new/larger subtrees.
    reduce_NTs = params["BNF_GRAMMAR"].concat_NTs

    # Pre-load the target string.
    target = params["REVERSE_MAPPING_TARGET"]

    for idx, snippet_info in enumerate(solution):
        # Get current snippet.
        snippet = snippet_info[2]

        # Find current snippet info.
        NT = snippet_info[1]

        # Get indexes of the current snippet
        indexes = snippet_info[0]
        start, end = indexes[0], indexes[1]

        # Find if the snippet root (NT) exists anywhere in the
        # reduction NTs.
        if NT in reduce_NTs:

            for reduce in reduce_NTs[NT]:
                # Now we're searching for a specific subset of keys in the
                # snippets dictionary.

                # Generate list of only the desired Non Terminals.
                NTs = reduce[2]

                if len(NTs) == 1:
                    # This choice leads directly to the parent, check if parent
                    # snippet already exists.

                    # Child is current snippet.
                    child = [[snippet, trackers.snippets[snippet]]]

                    # Get key for new snippet.
                    key, start, end = generate_key_and_check(
                        start, end, reduce, child)

                    # Create a new node for the solution list.
                    new_entry = [indexes, reduce[1], key]

                    # Insert the current node into the solution.
                    if new_entry not in solution:
                        solution.insert(idx + 1, new_entry)

                else:
                    # Find the index of the snippet root in the current
                    # reduction production choice.
                    NT_locs = [i for i, x in enumerate(NTs) if x[0] == NT]

                    for loc in NT_locs:
                        # We want to check each possible reduction option.

                        # Set where the original snippet starts and ends on
                        # the target string.
                        if loc == 0:
                            # The current snippet is at the start of the
                            # reduction attempt.
                            pre, aft = None, end

                        elif start == 0 and loc != 0:
                            # The current snippet is at the start of the target
                            # string, but we are trying to reduce_trees it with
                            # something before it.
                            break

                        elif end == len(params["TARGET"]) and loc != NT_locs[-1]:
                            # The current snippet is at the end of the target
                            # string, but we are trying to reduce_trees it with
                            # something after it.
                            break

                        elif loc == len(NTs):
                            # The current snippet is at the end of the
                            # reduction attempt.
                            pre, aft = start, None

                        else:
                            # The current snippet is in the middle of the
                            # reduction attempt.
                            pre, aft = start, end

                        alt_cs = list(range(len(NTs)))

                        # Initialise a list of children to be reduced.
                        children = [[] for _ in range(len(NTs))]

                        # Set original snippet into children.
                        children[loc] = [snippet, trackers.snippets[snippet]]

                        curr_idx = solution.index(snippet_info)

                        # Step 1: reduce everything before the current loc.
                        for item in reversed(alt_cs[:loc]):

                            if NTs[item][1] == "T":
                                # This is a terminal, decrement by length of T.

                                # Check output of target string.
                                check = target[pre - len(NTs[item][0]): pre]

                                if check == NTs[item][0]:
                                    # We have a match.

                                    # Generate fake key for snippets dict.
                                    key = str([pre - len(NTs[item][0]), pre])

                                    # Create new tree from this terminal.
                                    T_tree = Tree(check, None)

                                    # Add to children.
                                    children[item] = [key, T_tree]

                                    # Decrement target string index.
                                    pre -= len(NTs[item][0])

                                else:
                                    # No match.
                                    break

                            else:
                                # This is a NT. Check solution list for
                                # matching node.
                                available = [
                                    sol
                                    for sol in solution[:curr_idx]
                                    if sol[1] == NTs[item][0] and sol[0][1] == pre
                                ]

                                for check in available:
                                    # We have a match.

                                    # Set the correct child in our
                                    # children.
                                    children[item] = [
                                        check[2],
                                        trackers.snippets[check[2]],
                                    ]

                                    # Decrement target string index.
                                    child_len = get_num_from_str(check[2])
                                    pre -= child_len[1] - child_len[0]

                                    break

                        # Step 2: reduce everything after the loc.
                        for i, item in enumerate(alt_cs[loc + 1:]):

                            if NTs[item][1] == "T":
                                # This is a terminal, decrement by length of T.

                                # Check output of target string.
                                check = target[aft: aft + len(NTs[item][0])]

                                if check == NTs[item][0]:
                                    # We have a match.

                                    # Generate fake key for snippets dict.
                                    key = str([aft, aft + len(NTs[item][0])])

                                    # Create new tree from this terminal.
                                    T_tree = Tree(check, None)

                                    # Add to children.
                                    children[item] = [key, T_tree]

                                    # Increment target string index.
                                    aft += len(NTs[item][0])

                                else:
                                    # No match.
                                    break

                            else:
                                # We haven't looked ahead in the string,
                                # we can't add things we don't know yet.
                                break

                        if all([child != [] for child in children]):
                            # We have expanded all children and can collapse
                            # a node.

                            key, pre, aft = generate_key_and_check(
                                pre, aft, reduce, children
                            )

                            # Create a new node for the solution list.
                            new_entry = [[pre, aft], reduce[1], key]

                            # Add the new reduced entry to the solution.
                            if new_entry not in solution:
                                solution.insert(idx + 1, new_entry)


def parse_target_string():
    """
    Takes a list of terminal nodes and iteratively reduces that list until
    the solution has been found.

    :return: The complete parsed solution in the form of a GE individual.
    """

    # Sort snippets keys to generate the initial solution list of terminals.
    solution = sorted(
        [
            [get_num_from_str(snippet), get_NT_from_str(snippet), snippet]
            for snippet in trackers.snippets.keys()
        ]
    )

    # Perform reduction on the solution list.
    reduce(solution)

    # Check snippets for solution
    ind = check_snippets_for_solution()

    return ind


def main():
    """
    Run all functions to parse a target string into a GE individual.

    :return: A GE individual.
    """

    # Ensure there is a target to parse.
    if not params["REVERSE_MAPPING_TARGET"]:
        s = (
            "scripts.GE_LR_Parser.main\n"
            "Error: No target string specified for parsing."
        )
        raise Exception(s)

    # Parse the terminals in the target string.
    parse_terminals(params["REVERSE_MAPPING_TARGET"])

    # Iterate over the solution list until the target string is parsed.
    solution = parse_target_string()

    # Check the mapping of the solution and all aspects to ensure it is valid.
    check_ind_from_parser(solution, params["REVERSE_MAPPING_TARGET"])

    return solution


if __name__ == "__main__":

    # Set parameters
    set_params(sys.argv[1:], create_files=False)

    # Print parsed GE genome.
    print("\nGenome:\n", params["SEED_INDIVIDUALS"][0].genome)



from os import path

import numpy as np

from autooc.algorithm.parameters import params


def get_Xy_train_test_separate(train_filename, test_filename, skip_header=0):
    """
    Read in training and testing data files, and split each into X
    (all columns up to last) and y (last column).

    :param train_filename: The file name of the training dataset.
    :param test_filename: The file name of the testing dataset.
    :param skip_header: The number of header lines to skip.
    :return: Parsed numpy arrays of training and testing input (x) and
    output (y) data.
    """

    if params["DATASET_DELIMITER"]:
        # Dataset delimiter has been explicitly specified.
        delimiter = params["DATASET_DELIMITER"]

    else:
        # Try to auto-detect the field separator (i.e. delimiter).
        f = open(train_filename)
        for line in f:
            if line.startswith("#") or len(line) < 2:
                # Skip excessively short lines or commented out lines.
                continue

            else:
                # Set the delimiter.
                if "\t" in line:
                    delimiter = "\t"
                    break
                elif "," in line:
                    delimiter = ","
                    break
                elif ";" in line:
                    delimiter = ";"
                    break
                elif ":" in line:
                    delimiter = ":"
                    break
                else:
                    print(
                        "utilities.fitness.get_data.get_Xy_train_test_separate\n"
                        "Warning: Dataset delimiter not found. "
                        "Defaulting to whitespace delimiter."
                    )
                    delimiter = " "
                    break
        f.close()

    # Read in all training data.
    train_Xy = np.genfromtxt(
        train_filename, skip_header=skip_header, delimiter=delimiter
    )

    try:
        # Separate out input (X) and output (y) data.
        train_X = train_Xy[:, :-1].transpose()  # all columns but last
        train_y = train_Xy[:, -1].transpose()  # last column

    except IndexError:
        s = (
            "utilities.fitness.get_data.get_Xy_train_test_separate\n"
            "Error: specified delimiter '%s' incorrectly parses training "
            "data." % delimiter
        )
        raise Exception(s)

    if test_filename:
        # Read in all testing data.
        test_Xy = np.genfromtxt(
            test_filename, skip_header=skip_header, delimiter=delimiter
        )

        # Separate out input (X) and output (y) data.
        test_X = test_Xy[:, :-1].transpose()  # all columns but last
        test_y = test_Xy[:, -1].transpose()  # last column

    else:
        test_X, test_y = None, None

    return train_X, train_y, test_X, test_y


def get_data(train, test):
    """
    Return the training and test data for the current experiment.

    :param train: The desired training dataset.
    :param test: The desired testing dataset.
    :return: The parsed data contained in the dataset files.
    """

    # Get the path to the training dataset.
    train_set = path.join("..", "datasets", train)

    if test:
        # Get the path to the testing dataset.
        test_set = path.join("..", "datasets", test)

    else:
        # There is no testing dataset used.
        test_set = None

    # Read in the training and testing datasets from the specified files.
    training_in, training_out, test_in, test_out = get_Xy_train_test_separate(
        train_set, test_set, skip_header=1
    )

    return training_in, training_out, test_in, test_out



def step(agents):
    """
    Runs a single generation of the evolutionary algorithm process
    """
    # Loop over all the agents and apply their generic methods in sequence
    for agent in agents:
        # Sense the environment
        agent.sense(agents)

        # Based on the values from the sensor perform action
        agent.act()

        # Update the state of the agent
        agent.update()

    return agents



import argparse
from operator import attrgetter


class SortingHelpFormatter(argparse.HelpFormatter):
    """
    Custom class for sorting the arguments of the arg parser for printing. When
    "--help" is called, arguments will be listed in alphabetical order. Without
    this custom class, arguments will be printed in the order in which they are
    defined.
    """

    def add_arguments(self, actions):
        actions = sorted(actions, key=attrgetter("option_strings"))
        super(SortingHelpFormatter, self).add_arguments(actions)


def parse_cmd_args(arguments):
    """
    Parser for command line arguments specified by the user. Specified command
    line arguments over-write parameter file arguments, which themselves
    over-write original values in the algorithm.parameters.params dictionary.

    The argument parser structure is set up such that each argument has the
    following information:

        dest: a valid key from the algorithm.parameters.params dictionary
        type: an expected type for the specified option (i.e. str, int, float)
        help: a string detailing correct usage of the parameter in question.

    Optional info:

        default: The default setting for this parameter.
        action : The action to be undertaken when this argument is called.

    NOTE: You cannot add a new parser argument and have it evaluate "None" for
    its value. All parser arguments are set to "None" by default. We filter
    out arguments specified at the command line by removing any "None"
    arguments. Therefore, if you specify an argument as "None" from the
    command line and you evaluate the "None" string to a None instance, then it
    will not be included in the eventual parameters.params dictionary. A
    workaround for this would be to leave "None" command line arguments as
    strings and to eval them at a later stage.

    :param arguments: Command line arguments specified by the user.
    :return: A dictionary of parsed command line arguments, along with a
    dictionary of newly specified command line arguments which do not exist
    in the params dictionary.
    """

    # Initialise parser
    parser = argparse.ArgumentParser(
        formatter_class=SortingHelpFormatter,
        usage=argparse.SUPPRESS,
        description="""Welcome to PonyGE2 - Help.
        The following are the available command line arguments. Please see
        src/algorithm/parameters.py for a more detailed explanation of each
        argument and its possible values.""",
        epilog="""To try out PonyGE2 from the command line simply navigate to
        the src directory and type: python ponyge.py.""",
    )

    parser._optionals.title = "PonyGE2 command-line usage"

    # Set up class for parsing list arguments.
    class ListAction(argparse.Action):
        """
        Class for parsing a given string into a list.
        """

        def __init__(self, option_strings, **kwargs):
            super(ListAction, self).__init__(option_strings, **kwargs)

        def __call__(self, parser, namespace, value, option_string=None):
            if type(eval(value)) != list or any([type(i) != int for i in eval(value)]):
                s = (
                    "utilities.algorithm.command_line_parser.ListAction\n"
                    "Error: parameter %s is not a valid genome.\n"
                    "       Value given: %s" % (option_string, value)
                )
                raise Exception(s)
            else:
                setattr(namespace, self.dest, eval(value))

    # Set up class for checking float arguments.
    class FloatAction(argparse.Action):
        """
        Class for checking a given float is within the range [0:1].
        """

        def __init__(self, option_strings, **kwargs):
            super(FloatAction, self).__init__(option_strings, **kwargs)

        def __call__(self, parser, namespace, value, option_string=None):
            if not 0 <= float(value) <= 1:
                s = (
                    "utilities.algorithm.command_line_parser.FloatAction\n"
                    "Error: parameter %s outside allowed range [0:1].\n"
                    "       Value given: %s" % (option_string, value)
                )
                raise Exception(s)
            else:
                setattr(namespace, self.dest, float(value))

    # Set up class for checking raw string arguments to catch "tab" inputs.
    class CatchTabStr(argparse.Action):
        """
        Class for checking raw string arguments to catch "tab" inputs.
        """

        def __init__(self, option_strings, **kwargs):
            super(CatchTabStr, self).__init__(option_strings, **kwargs)

        def __call__(self, parser, namespace, value, option_string=None):
            if repr(value) == repr("\\t"):
                value = "\t"
            setattr(namespace, self.dest, value)

    # LOAD PARAMETERS FILE
    parser.add_argument(
        "--parameters",
        dest="PARAMETERS",
        type=str,
        help="Specifies the parameters file to be used. Must "
        "include the full file extension. Full file path"
        "does NOT need to be specified.",
    )

    # LOAD STEP AND SEARCH LOOP FUNCTIONS
    parser.add_argument(
        "--search_loop",
        dest="SEARCH_LOOP",
        type=str,
        help="Sets the desired search loop function.",
    )
    parser.add_argument(
        "--step", dest="STEP", type=str, help="Sets the desired search step function."
    )

    # POPULATION OPTIONS
    parser.add_argument(
        "--population_size",
        dest="POPULATION_SIZE",
        type=int,
        help="Sets the population size, requires int value.",
    )
    parser.add_argument(
        "--generations",
        dest="GENERATIONS",
        type=int,
        help="Sets the number of generations, requires int " "value.",
    )
    parser.add_argument(
        "--hill_climbing_history",
        dest="HILL_CLIMBING_HISTORY",
        type=int,
        help="Sets the history-length for late-acceptance"
        "and step-counting hill-climbing.",
    )
    parser.add_argument(
        "--schc_count_method",
        dest="SCHC_COUNT_METHOD",
        type=str,
        help="Sets the counting method for step-counting "
        'hill-climbing. Optional values are "count_all", '
        '"acp", and "imp".',
    )

    # INDIVIDUAL SIZE
    parser.add_argument(
        "--max_tree_depth",
        dest="MAX_TREE_DEPTH",
        type=int,
        help="Sets the max derivation tree depth for the "
        "algorithm, requires int value. The default max "
        "tree depth is set to None, i.e. trees can grow"
        "indefinitely. This can also be set by "
        "specifying the max tree depth to be 0.",
    )
    parser.add_argument(
        "--max_tree_nodes",
        dest="MAX_TREE_NODES",
        type=int,
        help="Sets the max derivation tree nodes for the "
        "algorithm, requires int value. The default max "
        "tree nodes is set to None, i.e. trees can grow"
        "indefinitely. This can also be set by "
        "specifying the max tree nodes to be 0.",
    )
    parser.add_argument(
        "--codon_size",
        dest="CODON_SIZE",
        type=int,
        help="Sets the range from 0 to codon_size to be used "
        "in genome, requires int value",
    )
    parser.add_argument(
        "--max_genome_length",
        dest="MAX_GENOME_LENGTH",
        type=int,
        help="Sets the maximum chromosome length for the "
        "algorithm, requires int value. The default max "
        "genome length is set to None, i.e. genomes can "
        "grow indefinitely. This can also be set by "
        "specifying the max genome length to be 0.",
    )
    parser.add_argument(
        "--max_wraps",
        dest="MAX_WRAPS",
        type=int,
        help="Sets the maximum number of times the genome "
        "mapping process can wrap over the length of the "
        "genome. Requires int value.",
    )
    parser.add_argument(
        "--permutation_ramps",
        dest="PERMUTATION_RAMPS",
        type=int,
        help="Set the number of depths permutations are "
        "calculated for (starting from the minimum path "
        "of the grammar). Mainly for use with "
        "the grammar analyser script. Requires int "
        "value.",
    )

    # INITIALISATION
    parser.add_argument(
        "--max_init_tree_depth",
        dest="MAX_INIT_TREE_DEPTH",
        type=int,
        help="Sets the max tree depth for initialisation.",
    )
    parser.add_argument(
        "--min_init_tree_depth",
        dest="MIN_INIT_TREE_DEPTH",
        type=int,
        help="Sets the min tree depth for initialisation.",
    )
    parser.add_argument(
        "--init_genome_length",
        dest="INIT_GENOME_LENGTH",
        type=int,
        help="Sets the length for chromosomes to be "
        "initialised to. Requires int value.",
    )
    parser.add_argument(
        "--initialisation",
        dest="INITIALISATION",
        type=str,
        help="Sets the initialisation strategy, requires a "
        'string such as "rhh" or a direct path string '
        'such as "operators.initialisation.rhh".',
    )

    # SELECTION
    parser.add_argument(
        "--selection",
        dest="SELECTION",
        type=str,
        help="Sets the selection to be used, requires string "
        'such as "tournament" or direct path string such '
        'as "operators.selection.tournament".',
    )
    parser.add_argument(
        "--invalid_selection",
        dest="INVALID_SELECTION",
        action="store_true",
        default=None,
        help="Allow for the selection of invalid individuals " "during selection.",
    )
    parser.add_argument(
        "--tournament_size",
        dest="TOURNAMENT_SIZE",
        type=int,
        help="Sets the number of individuals to contest tournament," " requires int.",
    )
    parser.add_argument(
        "--selection_proportion",
        dest="SELECTION_PROPORTION",
        action=FloatAction,
        help="Sets the proportion for truncation selection, "
        "requires float, e.g. 0.5.",
    )

    # OPERATOR OPTIONS
    parser.add_argument(
        "--within_used",
        dest="WITHIN_USED",
        default=None,
        action="store_true",
        help="Boolean flag for selecting whether or not "
        "mutation is confined to within the used portion "
        "of the genome. Default set to True.",
    )

    # CROSSOVER
    parser.add_argument(
        "--crossover",
        dest="CROSSOVER",
        type=str,
        help="Sets the type of crossover to be used, requires "
        'string such as "subtree" or direct path string '
        'such as "operators.crossover.subtree".',
    )
    parser.add_argument(
        "--crossover_probability",
        dest="CROSSOVER_PROBABILITY",
        action=FloatAction,
        help="Sets the crossover probability, requires float, " "e.g. 0.9.",
    )
    parser.add_argument(
        "--no_crossover_invalids",
        dest="NO_CROSSOVER_INVALIDS",
        default=None,
        action="store_true",
        help="Prevents invalid individuals from being " "generated by crossover.",
    )

    # MUTATION
    parser.add_argument(
        "--mutation",
        dest="MUTATION",
        type=str,
        help="Sets the type of mutation to be used, requires "
        'string such as "int_flip_per_codon" or direct '
        "path string such as "
        '"operators.mutation.int_flip_per_codon".',
    )
    parser.add_argument(
        "--mutation_events",
        dest="MUTATION_EVENTS",
        type=int,
        help="Sets the number of mutation events based on " "probability.",
    )
    parser.add_argument(
        "--mutation_probability",
        dest="MUTATION_PROBABILITY",
        action=FloatAction,
        help="Sets the rate of mutation probability for linear" " genomes",
    )
    parser.add_argument(
        "--no_mutation_invalids",
        dest="NO_MUTATION_INVALIDS",
        default=None,
        action="store_true",
        help="Prevents invalid individuals from being " "generated by mutation.",
    )

    # EVALUATION
    parser.add_argument(
        "--fitness_function",
        dest="FITNESS_FUNCTION",
        type=str,
        nargs="+",
        help="Sets the fitness function to be used. "
        'Requires string such as "regression". '
        "Multiple fitness functions can be specified"
        "for multiple objective optimisation (using "
        "NSGA-II). To specify multiple fitness "
        "functions simply enter in the desired names of"
        " the functions separated by spaces.",
    )
    parser.add_argument(
        "--dataset_train",
        dest="DATASET_TRAIN",
        type=str,
        help="For use with problems that use a dataset. "
        "Specifies the training data for evolution. "
        "Full file name must be specified.",
    )
    parser.add_argument(
        "--dataset_test",
        dest="DATASET_TEST",
        type=str,
        help="For use with problems that use a dataset. "
        "Specifies the testing data for evolution. "
        "Full file name must be specified.",
    )
    parser.add_argument(
        "--dataset_delimiter",
        dest="DATASET_DELIMITER",
        action=CatchTabStr,
        help="For use with problems that use a dataset. "
        "Specifies the delimiter for the dataset. "
        'Requires string such as "\\t".',
    )
    parser.add_argument(
        "--target",
        dest="TARGET",
        type=str,
        help="For string match problem. Requires target " "string.",
    )
    parser.add_argument(
        "--error_metric",
        dest="ERROR_METRIC",
        type=str,
        help="Sets the error metric to be used with supervised"
        " learning problems. Requires string such as "
        '"mse" or "rmse".',
    )
    parser.add_argument(
        "--optimize_constants",
        dest="OPTIMIZE_CONSTANTS",
        action="store_true",
        default=None,
        help="Whether to optimize numerical constants by "
        "gradient descent in supervised learning "
        "problems. Requires True or False, default "
        "False.",
    )
    parser.add_argument(
        "--multicore",
        dest="MULTICORE",
        action="store_true",
        default=None,
        help="Turns on multi-core evaluation.",
    )
    parser.add_argument(
        "--cores",
        dest="CORES",
        type=int,
        help="Specify the number of cores to be used for "
        "multi-core evaluation. Requires int.",
    )

    # REPLACEMENT
    parser.add_argument(
        "--replacement",
        dest="REPLACEMENT",
        type=str,
        help="Sets the replacement strategy, requires string "
        'such as "generational" or direct path string '
        'such as "operators.replacement.generational".',
    )
    parser.add_argument(
        "--elite_size",
        dest="ELITE_SIZE",
        type=int,
        help="Sets the number of elites to be used, requires " "int value.",
    )

    # PROBLEM SPECIFICS
    parser.add_argument(
        "--grammar_file",
        dest="GRAMMAR_FILE",
        type=str,
        help="Sets the grammar to be used, requires string.",
    )
    parser.add_argument(
        "--experiment_name",
        dest="EXPERIMENT_NAME",
        type=str,
        help="Optional parameter to save results in "
        "results/[EXPERIMENT_NAME] folder. If not "
        "specified then results are saved in default "
        "results folder.",
    )
    parser.add_argument(
        "--runs",
        dest="RUNS",
        type=int,
        help="Optional parameter to specify the number of "
        "runs to be performed for an experiment. Only "
        "used with experiment manager.",
    )
    parser.add_argument(
        "--extra_parameters",
        dest="EXTRA_PARAMETERS",
        type=str,
        nargs="+",
        help="Optional extra command line parameter for "
        "inclusion of any extra information required "
        "for user-specific runs. Can be whatever you "
        "want it to be. Specified arguments are parsed "
        "as a list. Specify as many values as desired, "
        "separated by spaces.",
    )

    # OPTIONS
    parser.add_argument(
        "--random_seed",
        dest="RANDOM_SEED",
        type=int,
        help="Sets the random seed to be used with both the "
        "standard Python RNG and the NumPy RNG. "
        "requires int value.",
    )
    parser.add_argument(
        "--debug",
        dest="DEBUG",
        action="store_true",
        default=None,
        help="Disables saving of all ancillary files.",
    )
    parser.add_argument(
        "--verbose",
        dest="VERBOSE",
        action="store_true",
        default=None,
        help="Turns on the verbose output of the program in "
        "terms of command line and extra files.",
    )
    parser.add_argument(
        "--silent",
        dest="SILENT",
        action="store_true",
        default=None,
        help="Prevents any output from being printed to the " "command line.",
    )
    parser.add_argument(
        "--save_all",
        dest="SAVE_ALL",
        action="store_true",
        default=None,
        help="Saves the best phenotypes at each generation.",
    )
    parser.add_argument(
        "--save_plots",
        dest="SAVE_PLOTS",
        action="store_true",
        default=None,
        help="Saves plots for best fitness.",
    )

    # REVERSE-MAPPING
    parser.add_argument(
        "--reverse_mapping_target",
        dest="REVERSE_MAPPING_TARGET",
        type=str,
        help="Target string to parse into a GE individual.",
    )
    parser.add_argument(
        "--target_seed_folder",
        dest="TARGET_SEED_FOLDER",
        type=str,
        help='Specify a target seed folder in the "seeds" '
        "directory that contains a population of "
        "individuals with which to seed a run.",
    )

    # STATE SAVING/LOADING
    parser.add_argument(
        "--save_state",
        dest="SAVE_STATE",
        action="store_true",
        default=None,
        help="Saves the state of the evolutionary run every "
        "generation. You can specify how often you want "
        "to save the state with the command "
        '"--save_state_step".',
    )
    parser.add_argument(
        "--save_state_step",
        dest="SAVE_STATE_STEP",
        type=int,
        help="Specifies how often the state of the current "
        "evolutionary run is saved (i.e. every n-th "
        "generation). Requires int value.",
    )
    parser.add_argument(
        "--load_state",
        dest="LOAD_STATE",
        type=str,
        help="Load an evolutionary run from a saved state. "
        "You must specify the full file path to the "
        "desired state file. Note that state files have "
        "no file type.",
    )

    # MULTIAGENT
    parser.add_argument(
        "--multiagent",
        dest="MULTIAGENT",
        action="store_true",
        default=None,
        help="This enable the multi-agent mode. If this mode is"
        " enabled the search_loop and step parameter are"
        " overridden with search_multiagent and step_multiagent"
        " respectively",
    )
    parser.add_argument(
        "--agent_size",
        dest="AGENT_SIZE",
        type=int,
        help="Specifies how many agents are initialize in"
        " the environment. By default 100 agents are initialize."
        " Greater the number of agents the time to find the"
        " would be reduced",
    )
    parser.add_argument(
        "--interaction_probability",
        dest="INTERACTION_PROBABILITY",
        action=FloatAction,
        help="Specifies the probability of agent interacting with"
        " other nearby agents in the environment. By default"
        " 0.5 probability is used. Higher the probability the time"
        " to find the solution would be reduced",
    )
    parser.add_argument(
        "--epochs",
        dest="EPOCHS",
        type=int,
        help="Specifies the number of epochs for autoencoder training",
    )

    # CACHING
    class CachingAction(argparse.Action):
        """
        Class for defining special mutually exclusive options for caching.
        """

        def __init__(
            self,
            option_strings,
            CACHE=None,
            LOOKUP_FITNESS=None,
            LOOKUP_BAD_FITNESS=None,
            MUTATE_DUPLICATES=None,
            **kwargs
        ):
            self.CACHE = CACHE
            self.LOOKUP_FITNESS = LOOKUP_FITNESS
            self.LOOKUP_BAD_FITNESS = LOOKUP_BAD_FITNESS
            self.MUTATE_DUPLICATES = MUTATE_DUPLICATES
            super(CachingAction, self).__init__(
                option_strings, nargs=0, **kwargs)

        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, "CACHE", self.CACHE)
            if (
                "LOOKUP_FITNESS" not in namespace
                or getattr(namespace, "LOOKUP_FITNESS") is not False
            ):
                # able to overwrite if True or None
                setattr(namespace, "LOOKUP_FITNESS", self.LOOKUP_FITNESS)
            if self.LOOKUP_BAD_FITNESS and "LOOKUP_BAD_FITNESS" not in namespace:
                setattr(namespace, "LOOKUP_BAD_FITNESS",
                        self.LOOKUP_BAD_FITNESS)
            if self.MUTATE_DUPLICATES and "MUTATE_DUPLICATES" not in namespace:
                setattr(namespace, "MUTATE_DUPLICATES", self.MUTATE_DUPLICATES)

    # Generate a mutually exclusive group for caching options. This means
    # that you cannot specify multiple caching options simultaneously,
    # only one at a time.
    parser.add_argument(
        "--cache",
        dest="CACHE",
        action=CachingAction,
        CACHE=True,
        LOOKUP_FITNESS=True,
        help="Tracks unique phenotypes and is used to " "lookup duplicate fitnesses.",
    )
    caching_group = parser.add_mutually_exclusive_group()
    caching_group.add_argument(
        "--dont_lookup_fitness",
        dest="CACHE",
        action=CachingAction,
        CACHE=True,
        LOOKUP_FITNESS=False,
        help="Uses cache to track duplicate "
        "individuals, but does not use the cache "
        "to save fitness evaluations.",
    )
    caching_group.add_argument(
        "--lookup_bad_fitness",
        dest="CACHE",
        action=CachingAction,
        CACHE=True,
        LOOKUP_FITNESS=False,
        LOOKUP_BAD_FITNESS=True,
        help="Gives duplicate phenotypes a bad fitness "
        "when encountered. Uses cache.",
    )
    caching_group.add_argument(
        "--mutate_duplicates",
        dest="CACHE",
        action=CachingAction,
        CACHE=True,
        LOOKUP_FITNESS=False,
        MUTATE_DUPLICATES=True,
        help="Replaces duplicate individuals with " "mutated versions. Uses cache.",
    )

    # Parse command line arguments using all above information.
    args, unknown = parser.parse_known_args(arguments)

    # All default args in the parser are set to "None". Only take arguments
    # which are not "None", i.e. arguments which have been passed in from
    # the command line.
    cmd_args = {key: value for key, value in vars(
        args).items() if value is not None}

    # Set "None" values correctly.
    for key in sorted(cmd_args.keys()):
        # Check all specified arguments.

        if type(cmd_args[key]) == str and cmd_args[key].lower() == "none":
            # Allow for people not using correct capitalisation.

            cmd_args[key] = None

    return cmd_args, unknown



from collections import deque

import numpy as np

from autooc.algorithm.parameters import params
from autooc.representation.tree import Tree
from autooc.utilities.representation.python_filter import python_filter


def mapper(genome, tree):
    """
    Wheel for mapping. Calls the correct mapper for a given _input. Checks
    the params dict to ensure the correct type of individual is being created.

    If a genome is passed in with no tree, all tree-related information is
    generated. If a tree is passed in with no genome, the genome is
    sequenced from the tree.

    :param genome: Genome of an individual.
    :param tree: Tree of an individual.
    :return: All components necessary for a fully mapped individual.
    """

    # one or other must be passed in, but not both
    assert genome or tree
    assert not (genome and tree)

    if genome:
        # We have a genome and need to map an individual from that genome.

        genome = list(genome)
        # This is a fast way of creating a new unique copy of the genome
        # (prevents cross-contamination of information between individuals).

        if params["GENOME_OPERATIONS"]:
            # Can generate tree information faster using
            # algorithm.mapper.map_ind_from_genome() if we don't need to
            # store the whole tree.
            (
                phenotype,
                genome,
                tree,
                nodes,
                invalid,
                depth,
                used_codons,
            ) = map_ind_from_genome(genome)

        else:
            # Build the tree using algorithm.mapper.map_tree_from_genome().
            (
                phenotype,
                genome,
                tree,
                nodes,
                invalid,
                depth,
                used_codons,
            ) = map_tree_from_genome(genome)

    else:
        # We have a tree.

        # genome, output, invalid, depth, and nodes can all be
        # generated by recursing through the tree once.

        genome, output, invalid, depth, nodes = tree.get_tree_info(
            params["BNF_GRAMMAR"].non_terminals.keys(), [], []
        )
        used_codons, phenotype = len(genome), "".join(output)

    if params["BNF_GRAMMAR"].python_mode and not invalid:
        # Grammar contains python code

        phenotype = python_filter(phenotype)

    if invalid:
        # Set values for invalid individuals.
        phenotype, nodes, depth, used_codons = None, np.NaN, np.NaN, np.NaN

    return phenotype, genome, tree, nodes, invalid, depth, used_codons


def map_ind_from_genome(genome):
    """
    A fast genotype to phenotype mapping process. Map input via rules to
    output. Does not require the recursive tree class, but still calculates
    tree information, e.g. number of nodes and maximum depth.

    :param genome: A genome to be mapped.
    :return: Output in the form of a phenotype string ('None' if invalid),
             Genome,
             None (this is reserved for the derivation tree),
             The number of nodes in the derivation,
             A boolean flag for whether or not the individual is invalid,
             The maximum depth of any node in the tree, and
             The number of used codons.
    """

    # Create local variables to avoid multiple dictionary lookups
    max_tree_depth, max_wraps = params["MAX_TREE_DEPTH"], params["MAX_WRAPS"]
    bnf_grammar = params["BNF_GRAMMAR"]

    n_input = len(genome)

    # Depth, max_depth, and nodes start from 1 to account for starting root
    # Initialise number of wraps at -1 (since
    used_input, current_depth, max_depth, nodes, wraps = 0, 1, 1, 1, -1

    # Initialise output as empty deque list (deque is a list-like container
    # with fast appends and pops on either end).
    output = deque()

    # Initialise the list of unexpanded non-terminals with the start rule.
    unexpanded_symbols = deque([(bnf_grammar.start_rule, 1)])

    while (wraps < max_wraps) and unexpanded_symbols:
        # While there are unexpanded non-terminals, and we are below our
        # wrapping limit, we can continue to map the genome.

        if max_tree_depth and (max_depth > max_tree_depth):
            # We have breached our maximum tree depth limit.
            break

        if (
            used_input % n_input == 0
            and used_input > 0
            and any([i[0]["type"] == "NT" for i in unexpanded_symbols])
        ):
            # If we have reached the end of the genome and unexpanded
            # non-terminals remain, then we need to wrap back to the start
            # of the genome again. Can break the while loop.
            wraps += 1

        # Expand a production from the list of unexpanded non-terminals.
        current_item = unexpanded_symbols.popleft()
        current_symbol, current_depth = current_item[0], current_item[1]

        if max_depth < current_depth:
            # Set the new maximum depth.
            max_depth = current_depth

        # Set output if it is a terminal.
        if current_symbol["type"] != "NT":
            output.append(current_symbol["symbol"])

        else:
            # Current item is a new non-terminal. Find associated production
            # choices.
            production_choices = bnf_grammar.rules[current_symbol["symbol"]]["choices"]
            no_choices = bnf_grammar.rules[current_symbol["symbol"]
                                           ]["no_choices"]

            # Select a production based on the next available codon in the
            # genome.
            current_production = genome[used_input % n_input] % no_choices

            # Use an input
            used_input += 1

            # Initialise children as empty deque list.
            children = deque()
            nt_count = 0

            for prod in production_choices[current_production]["choice"]:
                # iterate over all elements of chosen production rule.

                child = [prod, current_depth + 1]

                # Extendleft reverses the order, thus reverse adding.
                children.appendleft(child)
                if child[0]["type"] == "NT":
                    nt_count += 1

            # Add the new children to the list of unexpanded symbols.
            unexpanded_symbols.extendleft(children)

            if nt_count > 0:
                nodes += nt_count
            else:
                nodes += 1

    # Generate phenotype string.
    output = "".join(output)

    if len(unexpanded_symbols) > 0:
        # All non-terminals have not been completely expanded, invalid
        # solution.
        return None, genome, None, nodes, True, max_depth, used_input

    return output, genome, None, nodes, False, max_depth, used_input


def map_tree_from_genome(genome):
    """
    Maps a full tree from a given genome.

    :param genome: A genome to be mapped.
    :return: All components necessary for a fully mapped individual.
    """

    # Initialise an instance of the tree class
    tree = Tree(str(params["BNF_GRAMMAR"].start_rule["symbol"]), None)

    # Map tree from the given genome
    output, used_codons, nodes, depth, max_depth, invalid = genome_tree_map(
        tree, genome, [], 0, 0, 0, 0
    )

    # Build phenotype.
    phenotype = "".join(output)

    if invalid:
        # Return "None" phenotype if invalid
        return None, genome, tree, nodes, invalid, max_depth, used_codons

    else:
        return phenotype, genome, tree, nodes, invalid, max_depth, used_codons


def genome_tree_map(
    tree, genome, output, index, depth, max_depth, nodes, invalid=False
):
    """
    Recursive function which builds a tree using production choices from a
    given genome. Not guaranteed to terminate.

    :param tree: An instance of the representation.tree.Tree class.
    :param genome: A full genome.
    :param output: The list of all terminal nodes in a subtree. This is
    joined to become the phenotype.
    :param index: The index of the current location on the genome.
    :param depth: The current depth in the tree.
    :param max_depth: The maximum overall depth in the tree so far.
    :param nodes: The total number of nodes in the tree thus far.
    :param invalid: A boolean flag indicating whether or not the individual
    is invalid.
    :return: index, the index of the current location on the genome,
             nodes, the total number of nodes in the tree thus far,
             depth, the current depth in the tree,
             max_depth, the maximum overall depth in the tree,
             invalid, a boolean flag indicating whether or not the
             individual is invalid.
    """

    if not invalid and index < len(genome) * (params["MAX_WRAPS"] + 1):
        # If the solution is not invalid thus far, and if we still have
        # remaining codons in the genome, then we can continue to map the tree.

        if params["MAX_TREE_DEPTH"] and (max_depth > params["MAX_TREE_DEPTH"]):
            # We have breached our maximum tree depth limit.
            invalid = True

        # Increment and set number of nodes and current depth.
        nodes += 1
        depth += 1
        tree.id, tree.depth = nodes, depth

        # Find all production choices and the number of those production
        # choices that can be made by the current root non-terminal.
        productions = params["BNF_GRAMMAR"].rules[tree.root]["choices"]
        no_choices = params["BNF_GRAMMAR"].rules[tree.root]["no_choices"]

        # Set the current codon value from the genome.
        tree.codon = genome[index % len(genome)]

        # Select the index of the correct production from the list.
        selection = tree.codon % no_choices

        # Set the chosen production
        chosen_prod = productions[selection]

        # Increment the index
        index += 1

        # Initialise an empty list of children.
        tree.children = []

        for symbol in chosen_prod["choice"]:
            # Add children to the derivation tree by creating a new instance
            # of the representation.tree.Tree class for each child.

            if symbol["type"] == "T":
                # Append the child to the parent node. Child is a terminal, do
                # not recurse.
                tree.children.append(Tree(symbol["symbol"], tree))
                output.append(symbol["symbol"])

            elif symbol["type"] == "NT":
                # Append the child to the parent node.
                tree.children.append(Tree(symbol["symbol"], tree))

                # Recurse by calling the function again to map the next
                # non-terminal from the genome.
                output, index, nodes, d, max_depth, invalid = genome_tree_map(
                    tree.children[-1],
                    genome,
                    output,
                    index,
                    depth,
                    max_depth,
                    nodes,
                    invalid=invalid,
                )

    else:
        # Mapping incomplete, solution is invalid.
        return output, index, nodes, depth, max_depth, True

    # Find all non-terminals in the chosen production choice.
    NT_kids = [
        kid for kid in tree.children if kid.root in params["BNF_GRAMMAR"].non_terminals
    ]

    if not NT_kids:
        # There are no non-terminals in the chosen production choice, the
        # branch terminates here.
        depth += 1
        nodes += 1

    if not invalid:
        # The solution is valid thus far.

        if depth > max_depth:
            # Set the new maximum depth.
            max_depth = depth

        if params["MAX_TREE_DEPTH"] and (max_depth > params["MAX_TREE_DEPTH"]):
            # If our maximum depth exceeds the limit, the solution is invalid.
            invalid = True

    return output, index, nodes, depth, max_depth, invalid



import pickle
from copy import copy
from os import getcwd, makedirs, path
from shutil import rmtree

from autooc.algorithm.parameters import params
from autooc.utilities.stats import trackers


def save_stats_to_file(stats, end=False):
    """
    Write the results to a results file for later analysis

    :param stats: The stats.stats.stats dictionary.
    :param end: A boolean flag indicating whether or not the evolutionary
    process has finished.
    :return: Nothing.
    """

    if params["VERBOSE"]:
        filename = path.join(params["FILE_PATH"], "stats.tsv")
        savefile = open(filename, "a")
        for stat in sorted(stats.keys()):
            savefile.write(str(stats[stat]) + "\t")
        savefile.write("\n")
        savefile.close()

    elif end:
        filename = path.join(params["FILE_PATH"], "stats.tsv")
        savefile = open(filename, "a")
        for item in trackers.stats_list:
            for stat in sorted(item.keys()):
                savefile.write(str(item[stat]) + "\t")
            savefile.write("\n")
        savefile.close()


def save_stats_headers(stats):
    """
    Saves the headers for all stats in the stats dictionary.

    :param stats: The stats.stats.stats dictionary.
    :return: Nothing.
    """

    filename = path.join(params["FILE_PATH"], "stats.tsv")
    savefile = open(filename, "w")
    for stat in sorted(stats.keys()):
        savefile.write(str(stat) + "\t")
    savefile.write("\n")
    savefile.close()


def save_best_ind_to_file(stats, ind, end=False, name="best"):
    """
    Saves the best individual to a file.

    :param stats: The stats.stats.stats dictionary.
    :param ind: The individual to be saved to file.
    :param end: A boolean flag indicating whether or not the evolutionary
    process has finished.
    :param name: The name of the individual. Default set to "best".
    :return: Nothing.
    """

    filename = path.join(params["FILE_PATH"], (str(name) + ".txt"))
    savefile = open(filename, "w")
    savefile.write("Generation:\n" + str(stats["gen"]) + "\n\n")
    savefile.write("Phenotype:\n" + str(ind.phenotype) + "\n\n")
    savefile.write("Genotype:\n" + str(ind.genome) + "\n")
    savefile.write("Tree:\n" + str(ind.tree) + "\n")
    if hasattr(params["FITNESS_FUNCTION"], "training_test"):
        if end:
            savefile.write("\nTraining fitness:\n" + str(ind.training_fitness))
            savefile.write("\nTest fitness:\n" + str(ind.test_fitness))
        else:
            savefile.write("\nFitness:\n" + str(ind.fitness))
    else:
        savefile.write("\nFitness:\n" + str(ind.fitness))
    savefile.close()

    if params["ALWAYS_AT_HAND"] == True:
        makedirs(path.join(params["FILE_PATH"], str(name)), exist_ok=True)
        if ind.model.__class__.__name__ == "Sequential":
            ind.model.save(
                path.join(params["FILE_PATH"], str(name), "current_best_model"))
        elif ind.model.__class__.__name__ in ["OneClassSVM", "IsolationForest"]:
            with open(path.join(params["FILE_PATH"], str(name), "best_model.pkl"), 'wb') as f:
                print("SAVING NEW FILE\n\n\n\n\n\n\n\n\n")
                pickle.dump(ind.model, f)


def save_first_front_to_file(stats, end=False, name="first"):
    """
    Saves all individuals in the first front to individual files in a folder.

    :param stats: The stats.stats.stats dictionary.
    :param end: A boolean flag indicating whether or not the evolutionary
                process has finished.
    :param name: The name of the front folder. Default set to "first_front".
    :return: Nothing.
    """

    # Save the file path (we will be over-writing it).
    orig_file_path = copy(params["FILE_PATH"])

    # Define the new file path.
    params["FILE_PATH"] = path.join(orig_file_path, str(name) + "_front")

    # Check if the front folder exists already
    if path.exists(params["FILE_PATH"]):

        # Remove previous files.
        rmtree(params["FILE_PATH"])

    # Create front folder.
    makedirs(params["FILE_PATH"])

    for i, ind in enumerate(trackers.best_ever):
        # Save each individual in the first front to file.
        save_best_ind_to_file(stats, ind, end, name=str(i))

    # Re-set the file path.
    params["FILE_PATH"] = copy(orig_file_path)


def generate_folders_and_files():
    """
    Generates necessary folders and files for saving statistics and parameters.

    :return: Nothing.
    """

    if params["EXPERIMENT_NAME"]:
        # Experiment manager is being used.
        path_1 = path.join(params["RESULTS_PATH"], "results")

        if not path.isdir(path_1):
            # Create results folder.
            makedirs(path_1, exist_ok=True)

        # Set file path to include experiment name.
        params["FILE_PATH"] = path.join(path_1, params["EXPERIMENT_NAME"])

    else:
        # Set file path to results folder.
        params["FILE_PATH"] = path.join(params["RESULTS_PATH"], "results")

    # Generate save folders
    if not path.isdir(params["FILE_PATH"]):
        makedirs(params["FILE_PATH"], exist_ok=True)

    if not path.isdir(path.join(params["FILE_PATH"], str(params["TIME_STAMP"]))):
        makedirs(
            path.join(params["FILE_PATH"], str(params["TIME_STAMP"])), exist_ok=True
        )

    params["FILE_PATH"] = path.join(
        params["FILE_PATH"], str(params["TIME_STAMP"]))

    save_params_to_file()


def save_params_to_file():
    """
    Save evolutionary parameters in a parameters.txt file.

    :return: Nothing.
    """

    # Generate file path and name.
    filename = path.join(params["FILE_PATH"], "parameters.txt")
    savefile = open(filename, "w")

    # Justify whitespaces for pretty printing/saving.
    col_width = max(len(param) for param in params.keys())

    for param in sorted(params.keys()):

        # Create whitespace buffer for pretty printing/saving.
        spaces = [" " for _ in range(col_width - len(param))]
        savefile.write(str(param) + ": " + "".join(spaces) +
                       str(params[param]) + "\n")

    savefile.close()



#!/usr/bin/env python
import sys
from collections import Counter

import numpy as np
from sklearn.linear_model import ElasticNet, LinearRegression

from autooc.algorithm.fitness.get_data import get_data


def pprint(a, format_string="{0:.2f}"):
    """
    Function to pretty print a array without scientific notation and with given
    precision.

    Adapted from http://stackoverflow.com/a/18287838

    :param a: An input array.
    :param format_string: The desired precision level.
    :return: A formatted array.
    """

    return "[" + ", ".join(format_string.format(v, i) for i, v in enumerate(a)) + "]"


def fit_maj_class(train_X, train_y, test_X):
    """
    Use the majority class, for a binary problem...

    :param train_X: An array of input (X) training data.
    :param train_y: An array of expected output (Y) training data.
    :param test_X: An array of input (X) testint data.
    :return:
    """

    # Set training Y data to int type.
    train_y = train_y.astype(int)

    # Get all classes from training Y data, often just {0, 1} or {-1, 1}.
    classes = set(train_y)

    # Get majority class.
    maj = Counter(train_y).most_common(1)[0][0]

    # Generate model.
    model = "Majority class %d" % maj

    # Generate training and testing output values.
    yhat_train = maj * np.ones(len(train_y))
    yhat_test = maj * np.ones(len(test_y))

    return model, yhat_train, yhat_test


def fit_const(train_X, train_y, test_X):
    """
    Use the mean of the y training values as a predictor.

    :param train_X:
    :param train_y:
    :param test_X:
    :return:
    """
    mn = np.mean(train_y)
    yhat_train = np.ones(len(train_y)) * mn
    yhat_test = np.ones(len(test_y)) * mn
    model = "Const %.2f" % mn

    return model, yhat_train, yhat_test


def fit_lr(train_X, train_y, test_X):
    """
    Use linear regression to predict.

    :param train_X:
    :param train_y:
    :param test_X:
    :return:
    """
    lr = LinearRegression()
    lr.fit(train_X, train_y)
    yhat_train = lr.predict(train_X)
    yhat_test = lr.predict(test_X)
    model = "LR int %.2f coefs %s" % (lr.intercept_, pprint(lr.coef_))

    return model, yhat_train, yhat_test


def fit_enet(train_X, train_y, test_X):
    """
    Use linear regression to predict. Elastic net is LR with L1 and L2
    regularisation.

    :param train_X:
    :param train_y:
    :param test_X:
    :return:
    """
    enet = ElasticNet()
    enet.fit(train_X, train_y)
    model = "ElasticNet int %.2f coefs %s" % (
        enet.intercept_, pprint(enet.coef_))
    yhat_train = enet.predict(train_X)
    yhat_test = enet.predict(test_X)

    return model, yhat_train, yhat_test


if __name__ == "__main__":

    dataset_name = sys.argv[1]
    metric = sys.argv[2] if len(sys.argv) > 2 else "rmse"
    s = f"from .error_metric import {metric} as metric"
    exec(s)

    train_X, train_y, test_X, test_y = get_data(dataset_name)
    train_X = train_X.T
    test_X = test_X.T

    methods = [fit_maj_class, fit_const, fit_lr, fit_enet]
    for fit in methods:
        model, train_yhat, test_yhat = fit(train_X, train_y, test_X)
        error_train = metric(train_y, train_yhat)
        error_test = metric(test_y, test_yhat)
        print(
            "%s %s %s train error %.2f test error %.2f"
            % (metric.__name__, fit.__name__, model, error_train, error_test)
        )



"""Utilities for tracking progress of runs, including time taken per
generation, fitness plots, fitness caches, etc."""

cache = {}
# This dict stores the cache for an evolutionary run. The key for each entry
# is the phenotype of the individual, the value is its fitness.

runtime_error_cache = []
# This list stores a list of phenotypes which produce runtime errors over an
# evolutionary run.

best_fitness_list = []
# fitness_plot is simply a list of the best fitnesses at each generation.
# Useful for plotting evolutionary progress.

first_pareto_list = []
# first_pareto_list stores the list of all individuals stored on the first
# pareto front during multi objective optimisation.

time_list = []
# time_list stores the system time after each generation has been completed.
# Useful for keeping track of how long each generation takes.

stats_list = []
# List for storing stats at each generation
# Used when verbose mode is off to speed up program

best_ever = None
# Store the best ever individual here.



from math import floor
from os import getcwd, listdir, path
from random import randint, shuffle

from autooc.algorithm.parameters import params
from autooc.representation import individual
from autooc.representation.derivation import generate_tree, pi_grow
from autooc.representation.individual import Individual
from autooc.representation.latent_tree import latent_tree_random_ind
from autooc.representation.tree import Tree
from autooc.scripts import GE_LR_parser
from autooc.utilities.representation.python_filter import python_filter


def initialisation(size):
    """
    Perform selection on a population in order to select a population of
    individuals for variation.

    :param size: The size of the required population.
    :return: A full population generated using the specified initialisation
    technique.
    """

    # Decrease initialised population size by the number of seed individuals
    # (if any) to ensure that the total initial population size does not exceed
    # the limit.
    size -= len(params["SEED_INDIVIDUALS"])

    # Initialise empty population.
    individuals = params["INITIALISATION"](size)

    # Add seed individuals (if any) to current population.
    individuals.extend(params["SEED_INDIVIDUALS"])

    return individuals


def sample_genome():
    """
    Generate a random genome, uniformly.

    :return: A randomly generated genome.
    """
    genome = [
        randint(0, params["CODON_SIZE"]) for _ in range(params["INIT_GENOME_LENGTH"])
    ]
    return genome


def uniform_genome(size):
    """
    Create a population of individuals by sampling genomes uniformly.

    :param size: The size of the required population.
    :return: A full population composed of randomly generated individuals.
    """

    return [individual.Individual(sample_genome(), None) for _ in range(size)]


def uniform_tree(size):
    """
    Create a population of individuals by generating random derivation trees.

    :param size: The size of the required population.
    :return: A full population composed of randomly generated individuals.
    """

    return [generate_ind_tree(params["MAX_TREE_DEPTH"], "random") for _ in range(size)]


def seed_individuals(size):
    """
    Create a population of size where all individuals are copies of the same
    seeded individuals.

    :param size: The size of the required population.
    :return: A full population composed of the seeded individuals.
    """

    # Get total number of seed inds.
    no_seeds = len(params["SEED_INDIVIDUALS"])

    # Initialise empty population.
    individuals = []

    if no_seeds > 0:
        # A list of individuals has been specified as the seed.

        # Divide requested population size by the number of seeds.
        num_per_seed = floor(size / no_seeds)

        for ind in params["SEED_INDIVIDUALS"]:

            if not isinstance(ind, individual.Individual):
                # The seed object is not a PonyGE individual.
                s = (
                    "operators.initialisation.seed_individuals\n"
                    "Error: SEED_INDIVIDUALS instance is not a PonyGE "
                    "individual."
                )
                raise Exception(s)

            else:
                # Generate num_per_seed identical seed individuals.
                individuals.extend([ind.deep_copy()
                                   for _ in range(num_per_seed)])

        return individuals

    else:
        # No seed individual specified.
        s = (
            "operators.initialisation.seed_individuals\n"
            "Error: No seed individual specified for seed initialisation."
        )
        raise Exception(s)


def rhh(size):
    """
    Create a population of size using ramped half and half (or sensible
    initialisation) and return.

    :param size: The size of the required population.
    :return: A full population of individuals.
    """

    # Calculate the range of depths to ramp individuals from.
    depths = range(
        params["BNF_GRAMMAR"].min_ramp + 1, params["MAX_INIT_TREE_DEPTH"] + 1
    )
    population = []

    if size < 2:
        # If the population size is too small, can't use RHH initialisation.
        print("Error: population size too small for RHH initialisation.")
        print("Returning randomly built trees.")
        return [individual.Individual(sample_genome(), None) for _ in range(size)]

    elif not depths:
        # If we have no depths to ramp from, then params['MAX_INIT_DEPTH'] is
        # set too low for the specified grammar.
        s = (
            "operators.initialisation.rhh\n"
            "Error: Maximum initialisation depth too low for specified "
            "grammar."
        )
        raise Exception(s)

    else:
        if size % 2:
            # Population size is odd, need an even population for RHH
            # initialisation.
            size += 1
            print(
                "Warning: Specified population size is odd, "
                "RHH initialisation requires an even population size. "
                "Incrementing population size by 1."
            )

        if size / 2 < len(depths):
            # The population size is too small to fully cover all ramping
            # depths. Only ramp to the number of depths we can reach.
            depths = depths[: int(size / 2)]

        # Calculate how many individuals are to be generated by each
        # initialisation method.
        times = int(floor((size / 2) / len(depths)))
        remainder = int(size / 2 - (times * len(depths)))

        # Iterate over depths.
        for depth in depths:
            # Iterate over number of required individuals per depth.
            for i in range(times):

                # Generate individual using "Grow"
                ind = generate_ind_tree(depth, "random")

                # Append individual to population
                population.append(ind)

                # Generate individual using "Full"
                ind = generate_ind_tree(depth, "full")

                # Append individual to population
                population.append(ind)

        if remainder:
            # The full "size" individuals were not generated. The population
            # will be completed with individuals of random depths.
            depths = list(depths)
            shuffle(depths)

        for i in range(remainder):
            depth = depths.pop()

            # Generate individual using "Grow"
            ind = generate_ind_tree(depth, "random")

            # Append individual to population
            population.append(ind)

            # Generate individual using "Full"
            ind = generate_ind_tree(depth, "full")

            # Append individual to population
            population.append(ind)

        return population


def PI_grow(size):
    """
    Create a population of size using Position Independent Grow and return.

    :param size: The size of the required population.
    :return: A full population of individuals.
    """

    # Calculate the range of depths to ramp individuals from.
    depths = range(
        params["BNF_GRAMMAR"].min_ramp + 1, params["MAX_INIT_TREE_DEPTH"] + 1
    )
    population = []

    if size < 2:
        # If the population size is too small, can't use PI Grow
        # initialisation.
        print("Error: population size too small for PI Grow initialisation.")
        print("Returning randomly built trees.")
        return [individual.Individual(sample_genome(), None) for _ in range(size)]

    elif not depths:
        # If we have no depths to ramp from, then params['MAX_INIT_DEPTH'] is
        # set too low for the specified grammar.
        s = (
            "operators.initialisation.PI_grow\n"
            "Error: Maximum initialisation depth too low for specified "
            "grammar."
        )
        raise Exception(s)

    else:
        if size < len(depths):
            # The population size is too small to fully cover all ramping
            # depths. Only ramp to the number of depths we can reach.
            depths = depths[: int(size)]

        # Calculate how many individuals are to be generated by each
        # initialisation method.
        times = int(floor(size / len(depths)))
        remainder = int(size - (times * len(depths)))

        # Iterate over depths.
        for depth in depths:
            # Iterate over number of required individuals per depth.
            for i in range(times):

                # Generate individual using "Grow"
                ind = generate_PI_ind_tree(depth)

                # Append individual to population
                population.append(ind)

        if remainder:
            # The full "size" individuals were not generated. The population
            #  will be completed with individuals of random depths.
            depths = list(depths)
            shuffle(depths)

        for i in range(remainder):
            depth = depths.pop()

            # Generate individual using "Grow"
            ind = generate_PI_ind_tree(depth)

            # Append individual to population
            population.append(ind)

        return population


def generate_ind_tree(max_depth, method):
    """
    Generate an individual using a given subtree initialisation method.

    :param max_depth: The maximum depth for the initialised subtree.
    :param method: The method of subtree initialisation required.
    :return: A fully built individual.
    """

    # Initialise an instance of the tree class
    ind_tree = Tree(str(params["BNF_GRAMMAR"].start_rule["symbol"]), None)

    # Generate a tree
    genome, output, nodes, _, depth = generate_tree(
        ind_tree, [], [], method, 0, 0, 0, max_depth
    )

    # Get remaining individual information
    phenotype, invalid, used_cod = "".join(output), False, len(genome)

    if params["BNF_GRAMMAR"].python_mode:
        # Grammar contains python code

        phenotype = python_filter(phenotype)

    # Initialise individual
    ind = individual.Individual(genome, ind_tree, map_ind=False)

    # Set individual parameters
    ind.phenotype, ind.nodes = phenotype, nodes
    ind.depth, ind.used_codons, ind.invalid = depth, used_cod, invalid

    # Generate random tail for genome.
    ind.genome = genome + [
        randint(0, params["CODON_SIZE"]) for _ in range(int(ind.used_codons / 2))
    ]

    return ind


def generate_PI_ind_tree(max_depth):
    """
    Generate an individual using a given Position Independent subtree
    initialisation method.

    :param max_depth: The maximum depth for the initialised subtree.
    :return: A fully built individual.
    """

    # Initialise an instance of the tree class
    ind_tree = Tree(str(params["BNF_GRAMMAR"].start_rule["symbol"]), None)

    # Generate a tree
    genome, output, nodes, depth = pi_grow(ind_tree, max_depth)

    # Get remaining individual information
    phenotype, invalid, used_cod = "".join(output), False, len(genome)

    if params["BNF_GRAMMAR"].python_mode:
        # Grammar contains python code

        phenotype = python_filter(phenotype)

    # Initialise individual
    ind = individual.Individual(genome, ind_tree, map_ind=False)

    # Set individual parameters
    ind.phenotype, ind.nodes = phenotype, nodes
    ind.depth, ind.used_codons, ind.invalid = depth, used_cod, invalid

    # Generate random tail for genome.
    ind.genome = genome + [
        randint(0, params["CODON_SIZE"]) for _ in range(int(ind.used_codons / 2))
    ]

    return ind


def load_population(target):
    """
    Given a target folder, read all files in the folder and load/parse
    solutions found in each file.

    :param target: A target folder stored in the "seeds" folder.
    :return: A list of all parsed individuals stored in the target folder.
    """

    # Set path for seeds folder
    path_1 = path.join(getcwd(), "..", "seeds")

    if not path.isdir(path_1):
        # Seeds folder does not exist.

        s = (
            "scripts.seed_PonyGE2.load_population\n"
            "Error: `seeds` folder does not exist in root directory."
        )
        raise Exception(s)

    path_2 = path.join(path_1, target)

    if not path.isdir(path_2):
        # Target folder does not exist.

        s = (
            "scripts.seed_PonyGE2.load_population\n"
            "Error: target folder " + target + " does not exist in seeds directory."
        )
        raise Exception(s)

    # Get list of all target individuals in the target folder.
    target_inds = [i for i in listdir(path_2) if i.endswith(".txt")]

    # Initialize empty list for seed individuals.
    seed_inds = []

    for ind in target_inds:
        # Loop over all target individuals.

        # Get full file path.
        file_name = path.join(path_2, ind)

        # Initialise None data for ind info.
        genotype, phenotype = None, None

        # Open file.
        with open(file_name, "r") as f:

            # Read file.
            raw_content = f.read()

            # Read file.
            content = raw_content.split("\n")

            # Check if genotype is already saved in file.
            if "Genotype:" in content:

                # Get index location of genotype.
                gen_idx = content.index("Genotype:") + 1

                # Get the genotype.
                try:
                    genotype = eval(content[gen_idx])
                except:
                    s = (
                        "scripts.seed_PonyGE2.load_population\n"
                        "Error: Genotype from file "
                        + file_name
                        + " not recognized: "
                        + content[gen_idx]
                    )
                    raise Exception(s)

            # Check if phenotype (target string) is already saved in file.
            if "Phenotype:" in content:

                # Get index location of genotype.
                phen_idx = content.index("Phenotype:") + 1

                # Get the phenotype.
                phenotype = content[phen_idx]

                # TODO: Current phenotype is read in as single-line only. Split is performed on "\n", meaning phenotypes that span multiple lines will not be parsed correctly. This must be fixed in later editions.

            elif "Genotype:" not in content:
                # There is no explicit genotype or phenotype in the target
                # file, read in entire file as phenotype.
                phenotype = raw_content

        if genotype:
            # Generate individual from genome.
            ind = Individual(genotype, None)

            if phenotype and ind.phenotype != phenotype:
                s = (
                    "scripts.seed_PonyGE2.load_population\n"
                    "Error: Specified genotype from file "
                    + file_name
                    + " doesn't map to same phenotype. Check the specified "
                    "grammar to ensure all is correct: " +
                    params["GRAMMAR_FILE"]
                )
                raise Exception(s)

        else:
            # Set target for GE LR Parser.
            params["REVERSE_MAPPING_TARGET"] = phenotype

            # Parse target phenotype.
            ind = GE_LR_parser.main()

        # Add new ind to the list of seed individuals.
        seed_inds.append(ind)

    return seed_inds


def LTGE_initialisation(size):
    """Initialise a population in the LTGE representation."""

    pop = []
    for _ in range(size):

        # Random genotype
        g, ph = latent_tree_random_ind(
            params["BNF_GRAMMAR"], params["MAX_TREE_DEPTH"])

        # wrap up in an Individual and fix up various Individual attributes
        ind = individual.Individual(g, None, False)

        ind.phenotype = ph

        # number of nodes is the number of decisions in the genome
        ind.nodes = ind.used_codons = len(g)

        # each key is the length of a path from root
        ind.depth = max(len(k) for k in g)

        # in LTGE there are no invalid individuals
        ind.invalid = False

        pop.append(ind)
    return pop


# Set ramping attributes for ramped initialisers.
PI_grow.ramping = True
rhh.ramping = True



from multiprocessing import Pool

from autooc.algorithm.parameters import params
from autooc.fitness.evaluation import evaluate_fitness
from autooc.operators.initialisation import initialisation
from autooc.stats.stats import get_stats, stats
from autooc.utilities.algorithm.initialise_run import pool_init
from autooc.utilities.stats import trackers


def search_loop():
    """
    This is a standard search process for an evolutionary algorithm. Loop over
    a given number of generations.

    :return: The final population after the evolutionary process has run for
    the specified number of generations.
    """

    if params["MULTICORE"]:
        # initialize pool once, if multi-core is enabled
        params["POOL"] = Pool(
            processes=params["CORES"], initializer=pool_init, initargs=(
                params,)
        )  # , maxtasksperchild=1)

    # Initialise population
    individuals = initialisation(params["POPULATION_SIZE"])

    # Evaluate initial population
    individuals = evaluate_fitness(individuals)

    # Generate statistics for run so far
    get_stats(individuals)

    # Traditional GE
    for generation in range(1, (params["GENERATIONS"] + 1)):
        stats["gen"] = generation

        # New generation
        individuals = params["STEP"](individuals)

    if params["MULTICORE"]:
        # Close the workers pool (otherwise they'll live on forever).
        params["POOL"].close()

    return individuals


def search_loop_from_state():
    """
    Run the evolutionary search process from a loaded state. Pick up where
    it left off previously.

    :return: The final population after the evolutionary process has run for
    the specified number of generations.
    """

    individuals = trackers.state_individuals

    if params["MULTICORE"]:
        # initialize pool once, if multi-core is enabled
        params["POOL"] = Pool(
            processes=params["CORES"], initializer=pool_init, initargs=(
                params,)
        )  # , maxtasksperchild=1)

    # Traditional GE
    for generation in range(stats["gen"] + 1, (params["GENERATIONS"] + 1)):
        stats["gen"] = generation

        # New generation
        individuals = params["STEP"](individuals)

    if params["MULTICORE"]:
        # Close the workers pool (otherwise they'll live on forever).
        params["POOL"].close()

    return individuals



from random import choice, randint, random, sample

from autooc.algorithm.parameters import params
from autooc.representation import individual
from autooc.representation.latent_tree import (latent_tree_crossover,
                                              latent_tree_repair)
from autooc.utilities.representation.check_methods import check_ind


def crossover(parents):
    """
    Perform crossover on a population of individuals. The size of the crossover
    population is defined as params['GENERATION_SIZE'] rather than params[
    'POPULATION_SIZE']. This saves on wasted evaluations and prevents search
    from evaluating too many individuals.

    :param parents: A population of parent individuals on which crossover is to
    be performed.
    :return: A population of fully crossed over individuals.
    """

    # Initialise an empty population.
    cross_pop = []

    while len(cross_pop) < params["GENERATION_SIZE"]:

        # Randomly choose two parents from the parent population.
        inds_in = sample(parents, 2)

        # Perform crossover on chosen parents.
        inds_out = crossover_inds(inds_in[0], inds_in[1])

        if inds_out is None:
            # Crossover failed.
            pass

        else:

            # Extend the new population.
            cross_pop.extend(inds_out)

    return cross_pop


def crossover_inds(parent_0, parent_1):
    """
    Perform crossover on two selected individuals.

    :param parent_0: Parent 0 selected for crossover.
    :param parent_1: Parent 1 selected for crossover.
    :return: Two crossed-over individuals.
    """

    # Create copies of the original parents. This is necessary as the
    # original parents remain in the parent population and changes will
    # affect the originals unless they are cloned.
    ind_0 = parent_0.deep_copy()
    ind_1 = parent_1.deep_copy()

    # Crossover cannot be performed on invalid individuals.
    if not params["INVALID_SELECTION"] and (ind_0.invalid or ind_1.invalid):
        s = (
            "operators.crossover.crossover\nError: invalid individuals "
            "selected for crossover."
        )
        raise Exception(s)

    # Perform crossover on ind_0 and ind_1.
    inds = params["CROSSOVER"](ind_0, ind_1)

    # Check each individual is ok (i.e. does not violate specified limits).
    checks = [check_ind(ind, "crossover") for ind in inds]

    if any(checks):
        # An individual violates a limit.
        return None

    else:
        # Crossover was successful, return crossed-over individuals.
        return inds


def variable_onepoint(p_0, p_1):
    """
    Given two individuals, create two children using one-point crossover and
    return them. A different point is selected on each genome for crossover
    to occur. Note that this allows for genomes to grow or shrink in
    size. Crossover points are selected within the used portion of the
    genome by default (i.e. crossover does not occur in the tail of the
    individual).

    :param p_0: Parent 0
    :param p_1: Parent 1
    :return: A list of crossed-over individuals.
    """

    # Get the chromosomes.
    genome_0, genome_1 = p_0.genome, p_1.genome

    # Uniformly generate crossover points.
    max_p_0, max_p_1 = get_max_genome_index(p_0, p_1)

    # Select unique points on each genome for crossover to occur.
    pt_0, pt_1 = randint(1, max_p_0), randint(1, max_p_1)

    # Make new chromosomes by crossover: these slices perform copies.
    if random() < params["CROSSOVER_PROBABILITY"]:
        c_0 = genome_0[:pt_0] + genome_1[pt_1:]
        c_1 = genome_1[:pt_1] + genome_0[pt_0:]
    else:
        c_0, c_1 = genome_0[:], genome_1[:]

    # Put the new chromosomes into new individuals.
    ind_0 = individual.Individual(c_0, None)
    ind_1 = individual.Individual(c_1, None)

    return [ind_0, ind_1]


def fixed_onepoint(p_0, p_1):
    """
    Given two individuals, create two children using one-point crossover and
    return them. The same point is selected on both genomes for crossover
    to occur. Crossover points are selected within the used portion of the
    genome by default (i.e. crossover does not occur in the tail of the
    individual).

    :param p_0: Parent 0
    :param p_1: Parent 1
    :return: A list of crossed-over individuals.
    """

    # Get the chromosomes.
    genome_0, genome_1 = p_0.genome, p_1.genome

    # Uniformly generate crossover points.
    max_p_0, max_p_1 = get_max_genome_index(p_0, p_1)

    # Select the same point on both genomes for crossover to occur.
    pt = randint(1, min(max_p_0, max_p_1))

    # Make new chromosomes by crossover: these slices perform copies.
    if random() < params["CROSSOVER_PROBABILITY"]:
        c_0 = genome_0[:pt] + genome_1[pt:]
        c_1 = genome_1[:pt] + genome_0[pt:]
    else:
        c_0, c_1 = genome_0[:], genome_1[:]

    # Put the new chromosomes into new individuals.
    ind_0 = individual.Individual(c_0, None)
    ind_1 = individual.Individual(c_1, None)

    return [ind_0, ind_1]


def fixed_twopoint(p_0, p_1):
    """
    Given two individuals, create two children using two-point crossover and
    return them. The same points are selected on both genomes for crossover
    to occur. Crossover points are selected within the used portion of the
    genome by default (i.e. crossover does not occur in the tail of the
    individual).

    :param p_0: Parent 0
    :param p_1: Parent 1
    :return: A list of crossed-over individuals.
    """

    genome_0, genome_1 = p_0.genome, p_1.genome

    # Uniformly generate crossover points.
    max_p_0, max_p_1 = get_max_genome_index(p_0, p_1)

    # Select the same points on both genomes for crossover to occur.
    a, b = randint(1, max_p_0), randint(1, max_p_1)
    pt_0, pt_1 = min([a, b]), max([a, b])

    # Make new chromosomes by crossover: these slices perform copies.
    if random() < params["CROSSOVER_PROBABILITY"]:
        c_0 = genome_0[:pt_0] + genome_1[pt_0:pt_1] + genome_0[pt_1:]
        c_1 = genome_1[:pt_0] + genome_0[pt_0:pt_1] + genome_1[pt_1:]
    else:
        c_0, c_1 = genome_0[:], genome_1[:]

    # Put the new chromosomes into new individuals.
    ind_0 = individual.Individual(c_0, None)
    ind_1 = individual.Individual(c_1, None)

    return [ind_0, ind_1]


def variable_twopoint(p_0, p_1):
    """
    Given two individuals, create two children using two-point crossover and
    return them. Different points are selected on both genomes for crossover
    to occur. Note that this allows for genomes to grow or shrink in size.
    Crossover points are selected within the used portion of the genome by
    default (i.e. crossover does not occur in the tail of the individual).

    :param p_0: Parent 0
    :param p_1: Parent 1
    :return: A list of crossed-over individuals.
    """

    genome_0, genome_1 = p_0.genome, p_1.genome

    # Uniformly generate crossover points.
    max_p_0, max_p_1 = get_max_genome_index(p_0, p_1)

    # Select the same points on both genomes for crossover to occur.
    a_0, b_0 = randint(1, max_p_0), randint(1, max_p_1)
    a_1, b_1 = randint(1, max_p_0), randint(1, max_p_1)
    pt_0, pt_1 = min([a_0, b_0]), max([a_0, b_0])
    pt_2, pt_3 = min([a_1, b_1]), max([a_1, b_1])

    # Make new chromosomes by crossover: these slices perform copies.
    if random() < params["CROSSOVER_PROBABILITY"]:
        c_0 = genome_0[:pt_0] + genome_1[pt_2:pt_3] + genome_0[pt_1:]
        c_1 = genome_1[:pt_2] + genome_0[pt_0:pt_1] + genome_1[pt_3:]
    else:
        c_0, c_1 = genome_0[:], genome_1[:]

    # Put the new chromosomes into new individuals.
    ind_0 = individual.Individual(c_0, None)
    ind_1 = individual.Individual(c_1, None)

    return [ind_0, ind_1]


def subtree(p_0, p_1):
    """
    Given two individuals, create two children using subtree crossover and
    return them. Candidate subtrees are selected based on matching
    non-terminal nodes rather than matching terminal nodes.

    :param p_0: Parent 0.
    :param p_1: Parent 1.
    :return: A list of crossed-over individuals.
    """

    def do_crossover(tree0, tree1, shared_nodes):
        """
        Given two instances of the representation.tree.Tree class (
        derivation trees of individuals) and a list of intersecting
        non-terminal nodes across both trees, performs subtree crossover on
        these trees.

        :param tree0: The derivation tree of individual 0.
        :param tree1: The derivation tree of individual 1.
        :param shared_nodes: The sorted list of all non-terminal nodes that are
        in both derivation trees.
        :return: The new derivation trees after subtree crossover has been
        performed.
        """

        # Randomly choose a non-terminal from the set of permissible
        # intersecting non-terminals.
        crossover_choice = choice(shared_nodes)

        # Find all nodes in both trees that match the chosen crossover node.
        nodes_0 = tree0.get_target_nodes([], target=[crossover_choice])
        nodes_1 = tree1.get_target_nodes([], target=[crossover_choice])

        # Randomly pick a node.
        t0, t1 = choice(nodes_0), choice(nodes_1)

        # Check the parents of both chosen subtrees.
        p0 = t0.parent
        p1 = t1.parent

        if not p0 and not p1:
            # Crossover is between the entire tree of both tree0 and tree1.

            return t1, t0

        elif not p0:
            # Only t0 is the entire of tree0.
            tree0 = t1

            # Swap over the subtrees between parents.
            i1 = p1.children.index(t1)
            p1.children[i1] = t0

            # Set the parents of the crossed-over subtrees as their new
            # parents. Since the entire tree of t1 is now a whole
            # individual, it has no parent.
            t0.parent = p1
            t1.parent = None

        elif not p1:
            # Only t1 is the entire of tree1.
            tree1 = t0

            # Swap over the subtrees between parents.
            i0 = p0.children.index(t0)
            p0.children[i0] = t1

            # Set the parents of the crossed-over subtrees as their new
            # parents. Since the entire tree of t0 is now a whole
            # individual, it has no parent.
            t1.parent = p0
            t0.parent = None

        else:
            # The crossover node for both trees is not the entire tree.

            # For the parent nodes of the original subtrees, get the indexes
            # of the original subtrees.
            i0 = p0.children.index(t0)
            i1 = p1.children.index(t1)

            # Swap over the subtrees between parents.
            p0.children[i0] = t1
            p1.children[i1] = t0

            # Set the parents of the crossed-over subtrees as their new
            # parents.
            t1.parent = p0
            t0.parent = p1

        return tree0, tree1

    def intersect(l0, l1):
        """
        Returns the intersection of two sets of labels of nodes of
        derivation trees. Only returns matching non-terminal nodes across
        both derivation trees.

        :param l0: The labels of all nodes of tree 0.
        :param l1: The labels of all nodes of tree 1.
        :return: The sorted list of all non-terminal nodes that are in both
        derivation trees.
        """

        # Find all intersecting elements of both sets l0 and l1.
        shared_nodes = l0.intersection(l1)

        # Find only the non-terminals present in the intersecting set of
        # labels.
        shared_nodes = [
            i for i in shared_nodes if i in params["BNF_GRAMMAR"].non_terminals
        ]

        return sorted(shared_nodes)

    if random() > params["CROSSOVER_PROBABILITY"]:
        # Crossover is not to be performed, return entire individuals.
        ind0 = p_1
        ind1 = p_0

    else:
        # Crossover is to be performed.

        if p_0.invalid:
            # The individual is invalid.
            tail_0 = []

        else:
            # Save tail of each genome.
            tail_0 = p_0.genome[p_0.used_codons:]

        if p_1.invalid:
            # The individual is invalid.
            tail_1 = []

        else:
            # Save tail of each genome.
            tail_1 = p_1.genome[p_1.used_codons:]

        # Get the set of labels of non terminals for each tree.
        labels1 = p_0.tree.get_node_labels(set())
        labels2 = p_1.tree.get_node_labels(set())

        # Find overlapping non-terminals across both trees.
        shared_nodes = intersect(labels1, labels2)

        if len(shared_nodes) != 0:
            # There are overlapping NTs, cross over parts of trees.
            ret_tree0, ret_tree1 = do_crossover(
                p_0.tree, p_1.tree, shared_nodes)

        else:
            # There are no overlapping NTs, cross over entire trees.
            ret_tree0, ret_tree1 = p_1.tree, p_0.tree

        # Initialise new individuals using the new trees.
        ind0 = individual.Individual(None, ret_tree0)
        ind1 = individual.Individual(None, ret_tree1)

        # Preserve tails.
        ind0.genome = ind0.genome + tail_0
        ind1.genome = ind1.genome + tail_1

    return [ind0, ind1]


def get_max_genome_index(ind_0, ind_1):
    """
    Given two individuals, return the maximum index on each genome across
    which operations are to be performed. This can be either the used
    portion of the genome or the entire length of the genome.

    :param ind_0: Individual 0.
    :param ind_1: Individual 1.
    :return: The maximum index on each genome across which operations are to be
             performed.
    """

    if params["WITHIN_USED"]:
        # Get used codons range.

        if ind_0.invalid:
            # ind_0 is invalid. Default to entire genome.
            max_p_0 = len(ind_0.genome)

        else:
            max_p_0 = ind_0.used_codons

        if ind_1.invalid:
            # ind_1 is invalid. Default to entire genome.
            max_p_1 = len(ind_1.genome)

        else:
            max_p_1 = ind_1.used_codons

    else:
        # Get length of entire genome.
        max_p_0, max_p_1 = len(ind_0.genome), len(ind_1.genome)

    return max_p_0, max_p_1


def LTGE_crossover(p_0, p_1):
    """Crossover in the LTGE representation."""

    # crossover and repair.
    # the LTGE crossover produces one child, and is symmetric (ie
    # xover(p0, p1) is not different from xover(p1, p0)), but since it's
    # stochastic we can just run it twice to get two individuals
    # expected to be different.
    g_0, ph_0 = latent_tree_repair(
        latent_tree_crossover(p_0.genome, p_1.genome),
        params["BNF_GRAMMAR"],
        params["MAX_TREE_DEPTH"],
    )
    g_1, ph_1 = latent_tree_repair(
        latent_tree_crossover(p_0.genome, p_1.genome),
        params["BNF_GRAMMAR"],
        params["MAX_TREE_DEPTH"],
    )

    # wrap up in Individuals and fix up various Individual attributes
    ind_0 = individual.Individual(g_0, None, False)
    ind_1 = individual.Individual(g_1, None, False)

    ind_0.phenotype = ph_0
    ind_1.phenotype = ph_1

    # number of nodes is the number of decisions in the genome
    ind_0.nodes = ind_0.used_codons = len(g_0)
    ind_1.nodes = ind_1.used_codons = len(g_1)

    # each key is the length of a path from root
    ind_0.depth = max(len(k) for k in g_0)
    ind_1.depth = max(len(k) for k in g_1)

    # in LTGE there are no invalid individuals
    ind_0.invalid = False
    ind_1.invalid = False

    return [ind_0, ind_1]


# Set attributes for all operators to define linear or subtree representations.
variable_onepoint.representation = "linear"
fixed_onepoint.representation = "linear"
variable_twopoint.representation = "linear"
fixed_twopoint.representation = "linear"
subtree.representation = "subtree"
LTGE_crossover.representation = "latent tree"



import json
import subprocess
import sys
from os import path

from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff


class progsys(base_ff):
    """Fitness function for program synthesis problems. Grammars and datasets
    for 29 benchmark problems from doi.org/10.1145/2739480.2754769 are
    provided. Evaluation is done in a separate python process."""

    # constants required for formatting the code correctly
    INSERTCODE = "<insertCodeHere>"

    INDENTSPACE = "  "
    LOOPBREAK = "loopBreak"
    LOOPBREAKUNNUMBERED = "loopBreak%"
    LOOPBREAK_INITIALISE = "loopBreak% = 0"
    LOOPBREAK_IF = "if loopBreak% >"
    LOOPBREAK_INCREMENT = "loopBreak% += 1"
    FORCOUNTER = "forCounter"
    FORCOUNTERUNNUMBERED = "forCounter%"

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

        self.training, self.test, self.embed_header, self.embed_footer = self.get_data(
            params["DATASET_TRAIN"], params["DATASET_TEST"], params["GRAMMAR_FILE"]
        )
        self.eval = self.create_eval_process()
        if params["MULTICORE"]:
            print(
                "Warming: multi-core is not supported with progsys "
                "as fitness function.\n"
                "Fitness function only allows sequential evaluation."
            )

    def evaluate(self, ind, **kwargs):

        dist = kwargs.get("dist", "training")

        program = self.format_program(
            ind.phenotype, self.embed_header, self.embed_footer
        )
        data = self.training if dist == "training" else self.test
        program = "{}\n{}\n".format(data, program)
        eval_json = json.dumps(
            {
                "script": program,
                "timeout": 1.0,
                "variables": ["cases", "caseQuality", "quality"],
            }
        )

        self.eval.stdin.write((eval_json + "\n").encode())
        self.eval.stdin.flush()
        result_json = self.eval.stdout.readline()

        result = json.loads(result_json.decode())

        if "exception" in result and "JSONDecodeError" in result["exception"]:
            self.eval.stdin.close()
            self.eval = self.create_eval_process()

        if "quality" not in result:
            result["quality"] = sys.maxsize
        return result["quality"]

    @staticmethod
    def create_eval_process():
        """create separate python process for evaluation"""
        return subprocess.Popen(
            ["python", "scripts/python_script_evaluation.py"],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
        )

    def format_program(self, individual, header, footer):
        """formats the program by formatting the individual and adding
        a header and footer"""
        last_new_line = header.rindex("\n")
        indent = header[last_new_line + len("\n"): len(header)]
        return header + self.format_individual(individual, indent) + footer

    def format_individual(self, code, additional_indent=""):
        """format individual by adding appropriate indentation and loop break
        statements"""
        parts = code.split("\n")
        indent = 0
        string_builder = ""
        for_counter_number = 0
        first = True
        for part in parts:
            line = part.strip()
            # remove indentation if bracket is at the beginning of the line
            while line.startswith(":}"):
                indent -= 1
                line = line[2:].strip()

            # add indent
            if not first:
                string_builder += additional_indent
            else:
                first = False

            for i in range(0, indent):
                string_builder += self.INDENTSPACE

            # add indentation
            while line.endswith("{:"):
                indent += 1
                line = line[: len(line) - 2].strip()
            # remove indentation if bracket is at the end of the line
            while line.endswith(":}"):
                indent -= 1
                line = line[: len(line) - 2].strip()

            if self.LOOPBREAKUNNUMBERED in line:
                if self.LOOPBREAK_INITIALISE in line:
                    line = ""  # remove line
                elif self.LOOPBREAK_IF in line:
                    line = line.replace(
                        self.LOOPBREAKUNNUMBERED, self.LOOPBREAK)
                elif self.LOOPBREAK_INCREMENT in line:
                    line = line.replace(
                        self.LOOPBREAKUNNUMBERED, self.LOOPBREAK)
                else:
                    raise Exception("Python 'while break' is malformed.")
            elif self.FORCOUNTERUNNUMBERED in line:
                line = line.replace(
                    self.FORCOUNTERUNNUMBERED, self.FORCOUNTER +
                    str(for_counter_number)
                )
                for_counter_number += 1

            # add line to code
            string_builder += line
            string_builder += "\n"

        return string_builder

    def get_data(self, train, test, grammar):
        """Return the training and test data for the current experiment.
        A new get_data method is required to load from a sub folder and to
        read the embed file"""
        train_set = path.join("..", "datasets", "progsys", train)
        test_set = path.join("..", "datasets", "progsys", test)

        embed_file = path.join(
            "..", "grammars", "progsys", (grammar[8:-4] + "-Embed.txt")
        )
        with open(embed_file, "r") as embed:
            embed_code = embed.read()
        insert = embed_code.index(self.INSERTCODE)
        embed_header, embed_footer = "", ""
        if insert > 0:
            embed_header = embed_code[:insert]
            embed_footer = embed_code[insert + len(self.INSERTCODE):]
        with open(train_set, "r") as train_file, open(test_set, "r") as test_file:
            return train_file.read(), test_file.read(), embed_header, embed_footer



from autooc.fitness.evaluation import evaluate_fitness
from autooc.operators.crossover import crossover
from autooc.operators.mutation import mutation
from autooc.operators.replacement import replacement, steady_state
from autooc.operators.selection import selection
from autooc.stats.stats import get_stats


def step(individuals):
    """
    Runs a single generation of the evolutionary algorithm process:
        Selection
        Variation
        Evaluation
        Replacement

    :param individuals: The current generation, upon which a single
    evolutionary generation will be imposed.
    :return: The next generation of the population.
    """

    # Select parents from the original population.
    parents = selection(individuals)

    # Crossover parents and add to the new population.
    cross_pop = crossover(parents)

    # Mutate the new population.
    new_pop = mutation(cross_pop)

    # Evaluate the fitness of the new population.
    new_pop = evaluate_fitness(new_pop)

    # Replace the old population with the new population.
    individuals = replacement(new_pop, individuals)

    # Generate statistics for run so far
    get_stats(individuals)

    return individuals


def steady_state_step(individuals):
    """
    Runs a single generation of the evolutionary algorithm process,
    using steady state replacement.

    :param individuals: The current generation, upon which a single
    evolutionary generation will be imposed.
    :return: The next generation of the population.
    """

    individuals = steady_state(individuals)

    return individuals



import re

import scipy

from autooc.algorithm.parameters import params
from autooc.utilities.fitness.math_functions import *


def optimize_constants(x, y, ind):
    """
    Use gradient descent to search for values for the constants in
    ind.phenotype which minimise loss.

    :param x: Input (an array of x values).
    :param y: Expected output (expected y values for given inputs).
    :param ind: A GE individual.
    :return: The value of the error metric at those values
    """

    # Save the old phenotype (has c[0] etc, but may not be consecutive from
    # zero)
    ind.phenotype_original = ind.phenotype

    # Parse the phenotype to make the constants consecutive.
    s, n_consts = make_consts_consecutive(ind.phenotype)

    # Create new consecutive constant attribute for individual.
    ind.phenotype_consec_consts = s

    # Eval the phenotype.
    f = eval("lambda x, c: " + s)

    # Pre-load the error metric fitness function.
    loss = params["ERROR_METRIC"]

    if n_consts == 0:
        # ind doesn't refer to c: no need to optimize
        c = []
        fitness = loss(y, f(x, c))
        ind.opt_consts = c
        return fitness

    def obj(c):
        return loss(y, f(x, c))

    # obj is now a function of c only for L-BFGS-B. Using 0 as the init seems a
    # reasonable choice. But for scipy.curve_fit we might use [1.0] * n_consts.
    # Maybe other minimizers do better with some other choices? There are other
    # methods to try out.
    init = [0.0] * n_consts

    res = scipy.optimize.minimize(obj, init, method="L-BFGS-B")

    # the result is accessed like a dict
    ind.opt_consts = res["x"]  # the optimum values of the constants

    # the most useful form of the phenotype: c[0], c[1] etc replaced
    # with actual numbers, so can be eval'd directly
    ind.phenotype = replace_consts_with_values(s, ind.opt_consts)

    return res["fun"]


def make_consts_consecutive(s):
    """
    The given phenotype will have zero or more occurrences of each const c[0],
    c[1], etc. But eg it might have c[7], but no c[0]. We need to remap, eg:
        7 -> 0
        9 -> 1
    so that we just have c[0], c[1], etc.

    :param s: A given phenotype string.
    :return: The phenotype string but with consecutive constants.
    """

    p = r"c\[(\d+)\]"
    # find the consts, extract idxs as ints, unique-ify and sort
    const_idxs = sorted(map(int, set(re.findall(p, s))))

    for i, j in enumerate(const_idxs):
        ci = "c[%d]" % i
        cj = "c[%d]" % j
        s = s.replace(cj, ci)

    return s, len(const_idxs)


def replace_consts_with_values(s, c):
    """
    Replace the constants in a given string s with the values in a list c.

    :param s: A given phenotype string.
    :param c: A list of values which will replace the constants in the
    phenotype string.
    :return: The phenotype string with the constants replaced.
    """

    for i in range(len(c)):
        s = s.replace("c[%d]" % i, str(c[i]))

    return s



from collections import defaultdict

from numpy import isnan

from autooc.algorithm.parameters import params
from autooc.utilities.fitness.math_functions import percentile


def compute_pareto_metrics(population):
    """
    Compute the pareto fronts using NSGA-II.

    :param population: A population to be sorted into fronts using NSGA-II.
    :return: The pareto fronts.
    """

    # Calculate the pareto fronts using Non-Dominated Sorting.
    pareto = sort_non_dominated(population)

    # Calculate the crowding distance
    pareto = calculate_crowding_distance(pareto)

    return pareto


def sort_non_dominated(population):
    """Sort the first *k* *population* into different non-domination levels
    using the "Fast Nondominated Sorting Approach" proposed by Deb et al.,
    see [Deb2002]_. This algorithm has a time complexity of :math:`O(MN^2)`,
    where :math:`M` is the number of objectives and :math:`N` the number of
    individuals.

    :param population: A list of individuals to select from.

    :returns: A list of Pareto fronts (lists), the first list includes
              non-dominated individuals.

    .. [Deb2002] Deb, Pratab, Agarwal, and Meyarivan, "A fast elitist
       non-dominated sorting genetic algorithm for multi-objective
       optimization: NSGA-II", 2002.

    """

    # Initialise empty pareto class instance.
    pareto = ParetoInfo()

    # Compute the Inter-Quartile Range (+1) value used to normalize the
    # crowding distance
    pareto.compute_iqr(population)

    # The naming *p* and *q* is the same adopted in [Deb2002]_
    for p in population:

        # Compute the domination counter of p
        for q in population:

            if dominates(p, q):
                # Add *q* to the set of solutions dominated by *p*
                pareto.dominated_solutions[p].append(q)

            elif dominates(q, p):
                # Increment the domination counter of p
                pareto.update_domination_count(p, True)

        # *p* belongs to the first front
        if pareto.get_domination_count(p) == 0:
            pareto.fronts[0].append(p)
            pareto.rank[p] = 0

    # Initialize the front counter
    i = 0

    # Compute the fronts.
    while len(pareto.fronts[i]) > 0:

        # Used to store the members of the next front
        big_q = []

        # For each each solution in the current front
        for p in pareto.fronts[i]:

            # Decrement the domination counter of each solution dominated by
            # *p*
            for q in pareto.dominated_solutions[p]:
                pareto.update_domination_count(q, False)

                # If the counter reaches 0, the solution is added to the next
                # front.
                if pareto.get_domination_count(q) == 0:
                    pareto.rank[q] = i + 1
                    big_q.append(q)

        # Update the front counter
        i += 1

        # Add the solutions selected in this iteration to the next front
        pareto.fronts.append(big_q)

    return pareto


def dominates(individual1, individual2):
    """
    Returns whether or not *individual1* dominates *individual2*. An individual
    dominates another if all fitness values are at least as good on all
    objectives, and strictly better than on at least one objective.

    :param individual1: The individual that would be dominated.
    :param individual2: The individual dominant.
    :returns: :obj:`True` if *individual1* dominates *individual2*,
              :obj:`False` otherwise.
    """

    if any([isnan(fit) for fit in individual1.fitness]):
        # Individual 1 is invalid.
        return False

    elif any([isnan(fit) for fit in individual2.fitness]):
        # Individual 2 is invalid.
        return True

    # Get fitness functions.
    ffs = params["FITNESS_FUNCTION"].fitness_functions

    # Check how many fitness values are equal.
    equal_fit = [False] * len(ffs)

    # Iterate over all fitness values and fitness functions.
    for ind1_value, ind2_value, ff, i in zip(
        individual1.fitness, individual2.fitness, ffs, range(len(ffs))
    ):
        # Check for fitness equality
        if ind1_value == ind2_value:
            equal_fit[i] = True

        elif not compare_fitnesses(ind1_value, ind2_value, ff):
            # ind1 does not dominate over ind2.
            return False

    # If all fitness values are equal, no domination
    if all(equal_fit):
        return False

    return True


def compare_fitnesses(ind1_value, ind2_value, ff):
    """
    Comparison function for checking whether ind1 dominates ind2 on a given
    fitness value.

    :param ind1_value: The fitness of ind1.
    :param ind2_value: The fitness of ind2.
    :param ff: The fitness function that generated the above values.
    :return: Whether or not ind1_value is better than ind2_value.
    """

    if ff.maximise:
        # The fitness function is maximising.

        # Check whether ind1_value is better than ind2_value.
        return ind1_value > ind2_value

    else:
        # The fitness function is minimising.

        # Check whether ind1_value is better than ind2_value.
        return ind1_value < ind2_value


def calculate_crowding_distance(pareto):
    """
    Compute the crowding distance of each individual in each Pareto front.
    The value is stored inside the dictionary *crowding_distance* kept by
    the object *pareto*.

    :param pareto: A ParetoInfo object with the information regarding
                   the Pareto fronts defined by the current population

    :return: A list of Pareto fronts (lists), the first list includes
             non-dominated individuals.
    """

    # The crowding distance is computed per front
    for front in pareto.fronts:

        if len(front) > 0:

            # Number of solutions in the front
            solutions_num = len(front)

            for individual in front:
                # Initialize the distances
                pareto.crowding_distance[individual] = 0

            for m in range(pareto.n_objectives):
                # Sort the solutions using each objective value
                front = sorted(
                    front,
                    key=lambda item: params["FITNESS_FUNCTION"].value(
                        item.fitness, m),
                    reverse=params["FITNESS_FUNCTION"].fitness_functions[m].maximise,
                )

                # The boundary solutions are assigned an infinite distance
                # value
                pareto.crowding_distance[front[0]] = float("inf")
                pareto.crowding_distance[front[solutions_num - 1]
                                         ] = float("inf")

                # All other intermediate solutions have the distance computed
                for index in range(1, solutions_num - 1):
                    # The distance value equals to the absolute normalized
                    # difference in the function values of two adjacent
                    # solutions. The normalization uses (IQR + 1) instead of
                    # (max-min)
                    pareto.crowding_distance[front[index]] += (
                        params["FITNESS_FUNCTION"].value(
                            front[index + 1].fitness, m)
                        - params["FITNESS_FUNCTION"].value(front[index - 1].fitness, m)
                    ) / pareto.fitness_iqr[m]

    return pareto


def crowded_comparison_operator(self, other, pareto):
    """
    The crowded-comparison operator guides the selection process at the various
    stages of the algorithm toward a uniformly spread-out Pareto-optimal front.
    The operator returns True if *self* is better than *other* and
    False otherwise.

    :param self: First individual of the comparison
    :param other: Second individual of the comparison
    :param pareto: A ParetoInfo object with the information regarding
                   the Pareto fronts defined by the current population
    :return: True if *self* is better than *other* and False otherwise.
    """

    # Between two solutions with differing non-domination ranks, we prefer the
    # solution with the lower (better) rank. Otherwise, if both solutions
    # belong to the same front, then we prefer the solution that is located in
    # a lesser crowded region, i.e., with the larger crowding distance.
    if (pareto.rank[self] < pareto.rank[other]) or (
        pareto.rank[self] == pareto.rank[other]
        and pareto.crowding_distance[self] > pareto.crowding_distance[other]
    ):
        return True

    else:
        return False


def get_population_iqr(population, n_objectives):
    """
    Compute the inter-quartile range (IQR) of the population regarding
    each objective.

    :param population: The input population
    :param n_objectives: Total number of objectives
    :return: List with the IQR regarding each objective
    """

    # Initialise base IQR as 0 for each objective
    iqr = [0 for _ in range(n_objectives)]

    for m in range(n_objectives):
        # Iterate over all objectives

        # Sort the population with respect to the current objective.
        sorted_pop = sorted(
            population,
            key=lambda ind: params["FITNESS_FUNCTION"].value(ind.fitness, m),
            reverse=params["FITNESS_FUNCTION"].fitness_functions[m].maximise,
        )

        # Get the inter-quartile fitness ranges for the current objective.
        iqr[m] = params["FITNESS_FUNCTION"].value(
            percentile(sorted_pop, 75).fitness, m
        ) - params["FITNESS_FUNCTION"].value(percentile(sorted_pop, 25).fitness, m)
    return iqr


class ParetoInfo:
    def __init__(self):
        self.fronts = [[]]
        self.rank = dict()
        self.domination_count = dict()
        self.crowding_distance = dict()
        self.dominated_solutions = defaultdict(list)

        try:
            self.n_objectives = params["FITNESS_FUNCTION"].num_obj

        except AttributeError:
            s = (
                "utilities.algorithm.NSGA2\n"
                "Error: Specified fitness function does not have "
                "'num_obj' attribute.\n"
                "       If using multiple objective optimisation, ensure "
                "fitness.base_ff_classes.base_moo_ff is implemented.\n"
                "       See README documentation for more information."
            )
            raise Exception(s)

        self.fitness_iqr = [0] * self.n_objectives

    def compute_iqr(self, population):
        """
        Compute the Inter-Quartile Range for a population for all fitness
        objectives.

        :param population: A population.
        :return: Nothing.
        """

        # Get the inter-quartile ranges for all objectives.
        self.fitness_iqr = get_population_iqr(population, self.n_objectives)

        # If the IQR value is zero, we replace it for 1---which is equivalent
        # to disregard the normalization process for that objective dimension.
        self.fitness_iqr = [1 if i == 0 else i for i in self.fitness_iqr]

    def update_domination_count(self, individual, should_increment=True):
        """
        Update the domination count of the *individual* by incrementing:

            (*should_increment*=:obj:`True`)

        or decrementing:

            (*should_increment*=:obj:`False`)

        :param individual: The referring individual
        :param should_increment: Indicates if the methods increment or
                                 decrement the value.
        :return: Nothing.
        """

        if individual in self.domination_count:
            if should_increment:
                self.domination_count[individual] += 1
            else:
                self.domination_count[individual] -= 1
        else:
            if should_increment:
                self.domination_count[individual] = 1
            else:
                self.domination_count[individual] = -1

    def get_domination_count(self, individual):
        """
        Avoids references to uninitialised positions in the dictionary.

        :param individual: Individual used as key in the dictionary.
        :return: The value regarding the key, if any, or 0 otherwise.
        """

        if individual in self.domination_count:
            return self.domination_count[individual]

        return 0

    def get_crowding_distance(self, individual):
        """
        Avoids references to uninitialised positions in the dictionary.

        :param individual: Individual used as key in the dictionary.
        :return: The value regarding the key, if any, or 0 otherwise.
        """

        if individual in self.crowding_distance:
            return self.crowding_distance[individual]

        return 0



from autooc.fitness.base_ff_classes.base_ff import base_ff


class minimise_nodes(base_ff):
    """
    Fitness function class for minimising the number of nodes in a
    derivation tree.
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

    def evaluate(self, ind, **kwargs):
        return ind.nodes



from autooc.fitness.base_ff_classes.base_ff import base_ff


class pymax(base_ff):
    """
    Py-max is a max-style problem where the goal is to generate a function
    which outputs a large number. In the standard GP Max [Gathercole and
    Ross] problem this function can only use the constant (0.5) and functions
    (+, *). The Py-max problem allows more programming: numerical expressions,
    assignment statements and loops. See pymax.pybnf.

    Chris Gathercole and Peter Ross. An adverse interaction between crossover
    and restricted tree depth in genetic programming. In John R. Koza,
    David E. Goldberg, David B. Fogel, and Rick L. Riolo, editors, Genetic
    Programming 1996: Proceedings of the First Annual Conference.
    """

    maximise = True  # True as it ever was.

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

    def evaluate(self, ind, **kwargs):
        # ind.phenotype will be a string, including function definitions etc.
        # When we exec it, it will create a value XXX_output_XXX, but we exec
        # inside an empty dict for safety.

        p, d = ind.phenotype, {}

        # Exec the phenotype.
        exec(p, d)

        # Get the output
        s = d["XXX_output_XXX"]  # this is the program's output: a number.

        return s



from random import sample

from autooc.algorithm.parameters import params
from autooc.utilities.algorithm.NSGA2 import (compute_pareto_metrics,
                                             crowded_comparison_operator)


def selection(population):
    """
    Perform selection on a population in order to select a population of
    individuals for variation.

    :param population: input population
    :return: selected population
    """

    return params["SELECTION"](population)


def tournament(population):
    """
    Given an entire population, draw <tournament_size> competitors randomly and
    return the best. Only valid individuals can be selected for tournaments.

    :param population: A population from which to select individuals.
    :return: A population of the winners from tournaments.
    """

    # Initialise list of tournament winners.
    winners = []

    # The flag "INVALID_SELECTION" allows for selection of invalid individuals.
    if params["INVALID_SELECTION"]:
        available = population
    else:
        available = [i for i in population if not i.invalid]

    while len(winners) < params["GENERATION_SIZE"]:
        # Randomly choose TOURNAMENT_SIZE competitors from the given
        # population. Allows for re-sampling of individuals.
        competitors = sample(available, params["TOURNAMENT_SIZE"])

        # Return the single best competitor.
        winners.append(max(competitors))

    # Return the population of tournament winners.
    return winners


def truncation(population):
    """
    Given an entire population, return the best <proportion> of them.

    :param population: A population from which to select individuals.
    :return: The best <proportion> of the given population.
    """

    # Sort the original population.
    population.sort(reverse=True)

    # Find the cutoff point for truncation.
    cutoff = int(len(population) * float(params["SELECTION_PROPORTION"]))

    # Return the best <proportion> of the given population.
    return population[:cutoff]


def nsga2_selection(population):
    """Apply NSGA-II selection operator on the *population*. Usually, the
    size of *population* will be larger than *k* because any individual
    present in *population* will appear in the returned list at most once.
    Having the size of *population* equals to *k* will have no effect other
    than sorting the population according to their front rank. The
    list returned contains references to the input *population*. For more
    details on the NSGA-II operator see [Deb2002]_.

    :param population: A population from which to select individuals.
    :returns: A list of selected individuals.
    .. [Deb2002] Deb, Pratab, Agarwal, and Meyarivan, "A fast elitist
       non-dominated sorting genetic algorithm for multi-objective
       optimization: NSGA-II", 2002.
    """

    selection_size = params["GENERATION_SIZE"]
    tournament_size = params["TOURNAMENT_SIZE"]

    # Initialise list of tournament winners.
    winners = []

    # The flag "INVALID_SELECTION" allows for selection of invalid individuals.
    if params["INVALID_SELECTION"]:
        available = population
    else:
        available = [i for i in population if not i.invalid]

    # Compute pareto front metrics.
    pareto = compute_pareto_metrics(available)

    while len(winners) < selection_size:
        # Return the single best competitor.
        winners.append(pareto_tournament(available, pareto, tournament_size))

    return winners


def pareto_tournament(population, pareto, tournament_size):
    """
    The Pareto tournament selection uses both the pareto front of the
    individual and the crowding distance.

    :param population: A population from which to select individuals.
    :param pareto: The pareto front information.
    :param tournament_size: The size of the tournament.
    :return: The selected individuals.
    """

    # Initialise no best solution.
    best = None

    # Randomly sample *tournament_size* participants.
    participants = sample(population, tournament_size)

    for participant in participants:
        if best is None or crowded_comparison_operator(participant, best, pareto):
            best = participant

    return best


# Set attributes for all operators to define multi-objective operators.
nsga2_selection.multi_objective = True



from autooc.agent.agent import Agent
from autooc.algorithm.parameters import params
# from autooc.fitness.evaluation import evaluate_fitness
from autooc.stats.stats import stats

# from autooc.utilities.stats import trackers
# from autooc.operators.initialisation import initialisation
# from autooc.utilities.algorithm.initialise_run import pool_init


def create_agents(n, p):
    """
    Create a list of agent specified by n parameter
    """
    return [Agent(p) for a in range(n)]


def search_loop():
    """
    This loop is used when the multi-agent parameter is passed
    """
    # Create a list of agents based on the parameter passed
    agents = create_agents(params["AGENT_SIZE"],
                           params["INTERACTION_PROBABILITY"])

    # Multi-Agent based GE
    for generation in range(1, (params["GENERATIONS"] + 1)):
        stats["gen"] = generation

        # New generation
        agents = params["STEP"](agents)

    return [agent.individual[0] for agent in agents]



import json
import logging
import multiprocessing as mp
import sys
from queue import Empty
from types import ModuleType

logging.basicConfig(
    filename="python_log.txt",
    format="%(asctime)s:%(process)d:%(thread)d:%(message)s",
    level=logging.INFO,
)  # set to DEBUG for debug info ;)


class Worker(mp.Process):
    def __init__(self, consume, produce):
        super(Worker, self).__init__()
        self.consume = consume
        self.produce = produce
        self.stop = mp.Value("b", False)

    def run(self):
        # START LINUX: used to receive Memory Error faster in Linux
        try:
            import resource

            resource.setrlimit(
                resource.RLIMIT_AS, (2 ** 30, 2 ** 30)
            )  # 2 ** 30 == 1GB in bytes
        except ImportError:
            pass
        # END LINUX:
        while True:
            exception = None
            self.stop.value = False
            script = self.consume.get()
            if script:
                help_globals = {"stop": self.stop}
                try:
                    exec(script, help_globals)
                except BaseException as e:
                    exc_type, exc_obj, exc_tb = sys.exc_info()
                    exception = "{} {} {}".format(exc_type, exc_obj, e.args)
                if exception:
                    self.produce.put({"exception": exception})
                else:
                    self.produce.put(
                        {
                            key: value
                            for key, value in help_globals.items()
                            if not callable(value)
                            and not isinstance(  # cannot be a function
                                value, ModuleType
                            )
                            and key  # cannot be a module
                            not in ["__builtins__", "stop"]
                        }
                    )  # cannot be builtins or synchronized objects
                del help_globals
            else:
                break

    def stop_current(self):
        self.stop.value = True


if __name__ == "__main__":
    consume = mp.Queue()
    produce = mp.Queue()
    p = Worker(consume, produce)
    p.start()
    while True:
        try:
            message = input()
            logging.debug("Received input")
        except EOFError as err:
            # No data was read with input()
            # HeuristicLab is not running anymore
            # stop thread
            consume.put(None)
            if not p.join(5):
                p.terminate()
            break

        # do not trust user input
        try:
            message_dict = json.loads(message)
        except json.decoder.JSONDecodeError as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            exception = "{} {} {}".format(exc_type, exc_obj, e.args)
            logging.debug(exception)
            logging.debug(message)
            print(json.dumps({"exception": exception}), flush=True)
            continue
        consume.put(message_dict["script"])
        try:
            results = produce.get(block=True, timeout=message_dict["timeout"])
        except Empty:
            results = None
        if not results:
            p.stop_current()
            try:
                produce.get(block=True, timeout=message_dict["timeout"] * 10)
            except Empty:
                # START: Used to terminate worker process if it does not return
                # Possible reasons: OS X does not throw a MemoryError and might kill the worker itself
                #                   worker just takes too long in general
                p.terminate()
                consume = mp.Queue()
                produce = mp.Queue()
                p = Worker(consume, produce)
                p.start()
                logging.debug("terminated worker")
                # END:
            print(json.dumps({"exception": "Timeout occurred."}), flush=True)
            logging.debug("Sent output timeout")
        elif "exception" in results:
            print(json.dumps(results), flush=True)
            logging.debug("Sent output exception")
        else:
            ret_message_dict = {}
            for v in message_dict["variables"]:
                if v in results:
                    ret_message_dict[v] = (
                        list(results[v]) if isinstance(
                            results[v], set) else results[v]
                    )
            print(json.dumps(ret_message_dict), flush=True)
            logging.debug("Sent output normal")



from sys import version_info


def check_python_version():
    """
    Check the python version to ensure it is correct. PonyGE uses Python 3.

    :return: Nothing
    """

    if version_info.major < 3 or (version_info.minor < 5 and version_info.major == 3):
        s = (
            "\nError: Python version not supported.\n"
            "       Must use at least Python 3.5."
        )
        raise Exception(s)



from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff


class string_match(base_ff):
    """Fitness function for matching a string. Takes a string and returns
    fitness. Penalises output that is not the same length as the target.
    Penalty given to individual string components which do not match ASCII
    value of target."""

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

        # Set target string.
        self.target = params["TARGET"]

    def evaluate(self, ind, **kwargs):
        guess = ind.phenotype
        fitness = max(len(self.target), len(guess))
        # Loops as long as the shorter of two strings
        for (t_p, g_p) in zip(self.target, guess):
            if t_p == g_p:
                # Perfect match.
                fitness -= 1
            else:
                # Imperfect match, find ASCII distance to match.
                fitness -= 1 / (1 + (abs(ord(t_p) - ord(g_p))))
        return fitness



from autooc.algorithm.parameters import params
from autooc.fitness.evaluation import evaluate_fitness
from autooc.stats.stats import get_stats, stats
from autooc.utilities.stats import trackers

"""Hill-climbing is just about the simplest meta-heuristic there
is. It's of interest in GP/GE because of the lingering suspicion among
many researchers that crossover just doesn't work. This goes back to
90s work by Chellapilla and by O'Reilly. Today, many papers are
published which use mutation only.

Even a simple hill-climber may work well in GP/GE. However, the main
purpose of this module is to provide two modern hill-climbing methods
proposed by Bykov: late-acceptance hill-climbing (LAHC) and
step-counting hill-climbing (SCHC). Both reduce to simple
hill-climbing (HC) with a particular parameter choice, so this module
provides HC as a by-product.

Both LAHC and SCHC are implemented as search_loop-type functions. They
don't provide/require a step-style function. Hence, to use these, just
pass the appropriate dotted function name:

--search_loop algorithm.hill_climbing.LAHC_loop
--search_loop algorithm.hill_climbing.SCHC_loop


LAHC is a hill-climbing algorithm with a history mechanism. The
history mechanism is very simple (one extra parameter: the length of
the history) but in some domains it seems to provide a remarkable
performance improvement compared to hill-climbing itself and other
heuristics. It hasn't previously been used in GP/GE.

LAHC was proposed by Bykov [http://www.cs.nott.ac.uk/~yxb/LAHC/LAHC-TR.pdf].

In standard hill-climbing, where we accept a move to a new proposed
point (created by mutation) if that point is as good as or better than
the current point.

In LAHC, we accept the move if the new point is as good as or better
than that we encountered L steps ago (L for history length).

LAHC is not to be confused with an acceptance to the GECCO
late-breaking papers track.

Step-counting hill-climbing
[http://link.springer.com/article/10.1007/s10951-016-0469-x] is a
variant, proposed by Bykov as an improvement on LAHC. Although less
"natural" it may be slightly simpler to tune again. In SCHC, we
maintain a threshold cost value. We accept moves which are better than
that. We update it every L steps to the current cost value.

There are also two variants: instead of counting all steps, we can
count only accepted, or only improving moves.
"""


def LAHC_search_loop():
    """
    Search loop for Late Acceptance Hill Climbing.

    This is the LAHC pseudo-code from Burke and Bykov:

        Produce an initial solution best
        Calculate initial cost function C(best)
        Specify Lfa
        For all k in {0...Lfa-1} f_k := C(best)
        First iteration iters=0;
        Do until a chosen stopping condition
            Construct a candidate solution best*
            Calculate its cost function C(best*)
            idx := iters mod Lfa
            If C(best*)<=f_idx or C(best*)<=C(best)
            Then accept the candidate (best:=best*)
            Else reject the candidate (best:=best)
            Insert the current cost into the fitness array f_idx:=C(best)
            Increment the iteration number iters:=iters+1

    :return: The final population.
    """

    max_its = params["POPULATION_SIZE"] * params["GENERATIONS"]

    # Initialise population
    individuals = params["INITIALISATION"](params["POPULATION_SIZE"])

    # Evaluate initial population
    individuals = evaluate_fitness(individuals)

    # Generate statistics for run so far
    get_stats(individuals)

    # Find the best individual so far.
    best = trackers.best_ever

    # Set history.
    Lfa = params["HILL_CLIMBING_HISTORY"]
    history = [best for _ in range(Lfa)]

    # iters is the number of individuals examined so far.
    iters = len(individuals)

    for generation in range(1, (params["GENERATIONS"] + 1)):

        this_gen = []

        # even though there is no population, we will take account of
        # the pop size parameter: ie we'll save stats after every
        # "generation"
        for j in range(params["POPULATION_SIZE"]):

            this_gen.append(best)  # collect this "generation"

            # Mutate the best to get the candidate best
            candidate_best = params["MUTATION"](best)
            if not candidate_best.invalid:
                candidate_best.evaluate()

            # Find the index of the relevant individual from the late
            # acceptance history.
            idx = iters % Lfa

            if candidate_best >= history[idx]:
                best = candidate_best  # Accept the candidate

            else:
                pass  # reject the candidate

            # Set the new best into the history.
            history[idx] = best

            # Increment evaluation counter.
            iters += 1

            if iters >= max_its:
                # We have completed the total number of iterations.
                break

        # Get stats for this "generation".
        stats["gen"] = generation
        get_stats(this_gen)

        if iters >= max_its:
            # We have completed the total number of iterations.
            break

    return individuals


def SCHC_search_loop():
    """
    Search Loop for Step-Counting Hill-Climbing.

    This is the SCHC pseudo-code from Bykov and Petrovic.

        Produce an initial solution best
        Calculate an initial cost function C(best)
        Initial cost bound cost_bound := C(best)
        Initial counter counter := 0
        Specify history
        Do until a chosen stopping condition
            Construct a candidate solution best*
            Calculate the candidate cost function C(best*)
            If C(best*) < cost_bound or C(best*) <= C(best)
                Then accept the candidate best := best*
                Else reject the candidate best := best
            Increment the counter counter := counter + 1
            If counter >= history
                Then update the bound cost_bound := C(best)
                reset the counter counter := 0

        Two alternative counting methods (start at the first If):

        SCHC-acp counts only accepted moves:

            If C(best*) < cost_bound or C(best*) <= C(best)
                Then accept the candidate best := best*
                     increment the counter counter := counter + 1
                Else reject the candidate best := best
            If counter >= history
                Then update the bound cost_bound := C(best)
                     reset the counter counter := 0

        SCHC-imp counts only improving moves:

            If C(best*) < C(best)
                Then increment the counter counter := counter + 1
            If C(best*) < cost_bound or C(best*) <= C(best)
                Then accept the candidate best := best*
                Else reject the candidate best := best
            If counter >= history
                Then update the bound cost_bound := C(best)
                     reset the counter counter := 0

    :return: The final population.
    """

    # Calculate maximum number of evaluation iterations.
    max_its = params["POPULATION_SIZE"] * params["GENERATIONS"]
    count_method = params["SCHC_COUNT_METHOD"]

    # Initialise population
    individuals = params["INITIALISATION"](params["POPULATION_SIZE"])

    # Evaluate initial population
    individuals = evaluate_fitness(individuals)

    # Generate statistics for run so far
    get_stats(individuals)

    # Set best individual and initial cost bound.
    best = trackers.best_ever
    cost_bound = best.deep_copy()

    # Set history and counter.
    history = params["HILL_CLIMBING_HISTORY"]
    counter = 0

    # iters is the number of individuals examined/iterations so far.
    iters = len(individuals)

    for generation in range(1, (params["GENERATIONS"] + 1)):

        this_gen = []

        # even though there is no population, we will take account of
        # the pop size parameter: ie we'll save stats after every
        # "generation"
        for j in range(params["POPULATION_SIZE"]):

            this_gen.append(best)  # collect this "generation"

            # Mutate best to get candidate best.
            candidate_best = params["MUTATION"](best)
            if not candidate_best.invalid:
                candidate_best.evaluate()

            # count
            if count_method == "count_all":  # we count all iterations (moves)
                counter += 1  # increment the counter

            elif count_method == "acp":  # we count accepted moves only
                if candidate_best > cost_bound or candidate_best >= best:
                    counter += 1  # increment the counter

            elif count_method == "imp":  # we count improving moves only
                if candidate_best > best:
                    counter += 1  # increment the counter

            else:
                s = (
                    "algorithm.hill_climbing.SCHC_search_loop\n"
                    "Error: Unknown count method: %s" % (count_method)
                )
                raise Exception(s)

            # accept
            if candidate_best > cost_bound or candidate_best >= best:
                best = candidate_best  # accept the candidate

            else:
                pass  # reject the candidate

            if counter >= history:
                cost_bound = best  # update the bound
                counter = 0  # reset the counter

            # Increment iteration counter.
            iters += 1

            if iters >= max_its:
                # We have completed the total number of iterations.
                break

        # Get stats for this "generation".
        stats["gen"] = generation
        get_stats(this_gen)

        if iters >= max_its:
            # We have completed the total number of iterations.
            break

    return individuals



import pickle
import random
from os import path


def create_state(individuals):
    """
    Create a dictionary representing the current state of an evolutionary
    run. The state includes the current population, the current random state,
    the parameters dictionary, the stats dictionary, and all lists in the
    utilities.stats.trackers module.

    :param individuals: A population of individuals to be saved.
    :return: The complete state of a run.
    """

    from time import time

    from autooc.algorithm.parameters import params
    from autooc.stats.stats import stats
    from autooc.utilities.stats import trackers

    # Get time.
    state_time = time()

    # Get random state.
    random_state = random.getstate()

    # Create a picklable version of the params dictionary. Since the params
    # dictionary contains functions and class instances, we need to replace
    # these with the names of their respective modules, since module
    # instances are not picklable.
    pickle_params = {
        param: (check_name(params[param]) if callable(
            params[param]) else params[param])
        for param in params
    }

    # Create a picklable version of the trackers module.
    pickle_trackers = {
        i: getattr(trackers, i) for i in dir(trackers) if not i.startswith("__")
    }

    # Create state dictionary
    state = {
        "trackers": pickle_trackers,
        "params": pickle_params,
        "stats": stats,
        "individuals": individuals,
        "random_state": random_state,
        "time": state_time,
    }

    save_state(state)


def save_state(state):
    """
    Save the current state of a run. Allows for interrupted runs to be
    re-loaded and continued from the last save point.

    :param state: A dictionary describing the current state of a run.
    :return: Nothing.
    """

    # Create pickle file
    state_file = open(path.join(state["params"]["FILE_PATH"], "state"), "wb")

    # Save state information
    pickle.dump(state, state_file)

    # Close file.
    state_file.close()


def load_state(state):
    """
    Load in the state of a previous run.

    :param state: A .mat file containing all information about the state of
    a run.
    :return: The loaded state of a run.
    """

    # Open pickle file
    state_file = open(state, "rb")

    # Get state information
    loaded_state = pickle.load(state_file)

    # Close file.
    state_file.close()

    # Set state.
    individuals = set_state(loaded_state)

    # Return individuals.
    return individuals


def set_state(state):
    """
    Given a dictionary representing the state of an evolutionary run, set all
    aspects of the system to re-create that state. The state includes the
    current population, the current random state, the parameters dictionary,
    the stats dictionary, and all lists in the utilities.stats.trackers module.

    Sets all aspects of the system and then returns a population of
    individuals at the current generation.

    :param state: The complete state of a run.
    :return: A population of individuals.
    """

    from time import time

    from autooc.algorithm.parameters import params
    from autooc.stats.stats import stats
    from autooc.utilities.algorithm.initialise_run import set_param_imports
    from autooc.utilities.stats import trackers

    # Set random state.
    random.setstate(state["random_state"])

    # Set stats.
    for stat in state["stats"]:
        stats[stat] = state["stats"][stat]

    # Set trackers.
    for tracker in state["trackers"]:
        setattr(trackers, tracker, state["trackers"][tracker])

    # Set parameters.
    for param in state["params"]:
        params[param] = state["params"][param]

    # Set correct param imports for specified function options, including
    # error metrics and fitness functions.
    set_param_imports()

    # Set time adjustment to account for old time.
    stats["time_adjust"] = time() - state["time"]

    return state["individuals"]


def check_name(obj):
    """
    Function for returning the name of a callable object. Function and class
    instances are handled differently, so we use a try/except clause to
    differentiate between the two.

    :param obj: An object for which we want to find the name.
    :return: The name of the object
    """

    try:
        return obj.__name__
    except AttributeError:
        return obj.__class__.__name__



import numpy as np

np.seterr(all="raise")


class base_ff:
    """
    Base fitness function class from which all fitness functions inherit.

    This is an abstract class which exists just to be subclassed; it should not
    be instantiated.
    """

    # Default default fitness function is NaN.
    default_fitness = np.NaN

    # Default fitness objective is to minimise fitness.
    maximise = False

    def __init__(self):
        pass

    def __call__(self, ind, **kwargs):
        """


        :param ind: An individual to be evaluated.
        :return: The fitness of the evaluated individual.
        """

        try:
            # Evaluate the fitness using the evaluate() function. This function
            # can be over-written by classes which inherit from this base
            # class.
            fitness = self.evaluate(ind, **kwargs)

        except (FloatingPointError, ZeroDivisionError, OverflowError, MemoryError):
            # FP err can happen through eg overflow (lots of pow/exp calls)
            # ZeroDiv can happen when using unprotected operators
            fitness = base_ff.default_fitness

            # These individuals are valid (i.e. not invalids), but they have
            # produced a runtime error.
            ind.runtime_error = True

        except Exception as err:
            # Other errors should not usually happen (unless we have
            # an unprotected operator) so user would prefer to see them.
            print(err)
            raise

        return fitness

    def evaluate(self, ind, **kwargs):
        """
        Default fitness execution call for all fitness functions. When
        implementing a new fitness function, merely over-write this function
        in your own fitness function. All fitness functions must inherit from
        the base fitness function class.

        :param ind: An individual to be evaluated.
        :param kwargs: Optional extra arguments.
        :return: The fitness of the evaluated individual.
        """

        # Evaluate the fitness of the phenotype
        fitness = eval(ind.phenotype)

        return fitness



import time

import numpy as np
import pandas as pd
import tensorflow as tf

from autooc.algorithm.mapper import mapper
from autooc.algorithm.parameters import params


class Individual(object):
    """
    A GE individual.
    """

    def __init__(self, genome, ind_tree, map_ind=True):
        """
        Initialise an instance of the individual class (i.e. create a new
        individual).

        :param genome: An individual's genome.
        :param ind_tree: An individual's derivation tree, i.e. an instance
        of the representation.tree.Tree class.
        :param map_ind: A boolean flag that indicates whether or not an
        individual needs to be mapped.
        """

        if map_ind:
            # The individual needs to be mapped from the given input
            # parameters.
            (
                self.phenotype,
                self.genome,
                self.tree,
                self.nodes,
                self.invalid,
                self.depth,
                self.used_codons,
            ) = mapper(genome, ind_tree)

        else:
            # The individual does not need to be mapped.
            self.genome, self.tree = genome, ind_tree

        self.fitness = params["FITNESS_FUNCTION"].default_fitness
        self.runtime_error = False
        self.name = None,
        self.id = f"ID_{int(time.time()*1000000)}"
        self.training_time = 0
        self.predict_time = 0
        self.num_params = 0
        self.model = None
        self.threshold = 0
        self.losses = None
        self.rec_error = 0

    def __lt__(self, other):
        """
        Set the definition for comparison of two instances of the individual
        class by their fitness values. Allows for sorting/ordering of a
        population of individuals. Note that numpy NaN is used for invalid
        individuals and is used by some fitness functions as a default fitness.
        We implement a custom catch for these NaN values.

        :param other: Another instance of the individual class (i.e. another
        individual) with which to compare.
        :return: Whether or not the fitness of the current individual is
        greater than the comparison individual.
        """

        if np.isnan(self.fitness):
            return True
        elif np.isnan(other.fitness):
            return False
        else:
            return (
                self.fitness < other.fitness
                if params["FITNESS_FUNCTION"].maximise
                else other.fitness < self.fitness
            )

    def __le__(self, other):
        """
        Set the definition for comparison of two instances of the individual
        class by their fitness values. Allows for sorting/ordering of a
        population of individuals. Note that numpy NaN is used for invalid
        individuals and is used by some fitness functions as a default fitness.
        We implement a custom catch for these NaN values.

        :param other: Another instance of the individual class (i.e. another
        individual) with which to compare.
        :return: Whether or not the fitness of the current individual is
        greater than or equal to the comparison individual.
        """

        if np.isnan(self.fitness):
            return True
        elif np.isnan(other.fitness):
            return False
        else:
            return (
                self.fitness <= other.fitness
                if params["FITNESS_FUNCTION"].maximise
                else other.fitness <= self.fitness
            )

    def __str__(self):
        """
        Generates a string by which individuals can be identified. Useful
        for printing information about individuals.

        :return: A string describing the individual.
        """
        return "Individual: " + str(self.phenotype) + "; " + str(self.fitness)

    def deep_copy(self):
        """
        Copy an individual and return a unique version of that individual.

        :return: A unique copy of the individual.
        """

        if not params["GENOME_OPERATIONS"]:
            # Create a new unique copy of the tree.
            new_tree = self.tree.__copy__()

        else:
            new_tree = None

        # Create a copy of self by initialising a new individual.
        new_ind = Individual(self.genome.copy(), new_tree, map_ind=False)

        # Set new individual parameters (no need to map genome to new
        # individual).
        new_ind.phenotype, new_ind.invalid = self.phenotype, self.invalid
        new_ind.depth, new_ind.nodes = self.depth, self.nodes
        new_ind.used_codons = self.used_codons
        new_ind.runtime_error = self.runtime_error

        return new_ind

    def evaluate(self):
        """
        Evaluates phenotype in using the fitness function set in the params
        dictionary. For regression/classification problems, allows for
        evaluation on either training or test distributions. Sets fitness
        value.

        :return: Nothing unless multi-core evaluation is being used. In that
        case, returns self.
        """

        # Evaluate fitness using specified fitness function.
        self.fitness = params["FITNESS_FUNCTION"](self)

        if params["MULTICORE"]:
            return self

    def predict(self, X, anomaly_class, normal_class, threshold):
        #print(self.model.__class__.__name__)
        if self.model.__class__.__name__ == "Sequential":
            preds = self.predict_autoencoder(
                X, anomaly_class, normal_class, threshold)
        elif self.model.__class__.__name__ in ["OneClassSVM", "IsolationForest", "LocalOutlierFactor"]:
            preds = self.predict_sklearn(X, anomaly_class, normal_class)

        return preds

    def predict_autoencoder(self, X, anomaly_class, normal_class, threshold):
        reconstructions = self.reconstruct(X)
        loss = tf.keras.losses.mae(reconstructions, X)
        anomaly_mask = pd.Series(loss) > threshold
        preds = anomaly_mask.map(
            lambda x: anomaly_class if x == True else normal_class)
        #preds = tf.math.less(loss, self.threshold).numpy() * 1

        return preds

    def predict_sklearn(self, X, anomaly_class, normal_class):
        """_summary_

        Args:
            X (_type_): _description_
            anomaly_class (int, str): _description_
            normal_class (int, str): _description_

        Returns:
            _type_: _description_
        """

        preds = self.model.predict(X)
        preds = [(lambda x: anomaly_class if x == -1 else normal_class)(x)
                 for x in preds]

        return preds

    def reconstruct(self, X):
        """ Use the individual autoencoder to reconstruct the data.

        Args:
            X (np.array): data to reconstruct

        Returns:
            np.array: reconstructed dataset
        """
        reconstructions = self.model.predict(X)

        return reconstructions

    def get_id(self):
        return self.id



import numpy as np

from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff
from autooc.utilities.fitness.get_data import get_data
from autooc.utilities.fitness.math_functions import *
from autooc.utilities.fitness.optimize_constants import optimize_constants

np.seterr(all="raise")


class supervised_learning(base_ff):
    """
    Fitness function for supervised learning, ie regression and
    classification problems. Given a set of training or test data,
    returns the error between y (true labels) and yhat (estimated
    labels).

    We can pass in the error metric and the dataset via the params
    dictionary. Of error metrics, eg RMSE is suitable for regression,
    while F1-score, hinge-loss and others are suitable for
    classification.

    This is an abstract class which exists just to be subclassed:
    should not be instantiated.
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

        # Get training and test data
        self.training_in, self.training_exp, self.test_in, self.test_exp = get_data(
            params["DATASET_TRAIN"], params["DATASET_TEST"]
        )

        # Find number of variables.
        self.n_vars = np.shape(self.training_in)[0]

        # Regression/classification-style problems use training and test data.
        if params["DATASET_TEST"]:
            self.training_test = True

    def evaluate(self, ind, **kwargs):
        """
        Note that math functions used in the solutions are imported from either
        utilities.fitness.math_functions or called from numpy.

        :param ind: An individual to be evaluated.
        :param kwargs: An optional parameter for problems with training/test
        data. Specifies the distribution (i.e. training or test) upon which
        evaluation is to be performed.
        :return: The fitness of the evaluated individual.
        """

        dist = kwargs.get("dist", "training")

        if dist == "training":
            # Set training datasets.
            x = self.training_in
            y = self.training_exp

        elif dist == "test":
            # Set test datasets.
            x = self.test_in
            y = self.test_exp

        else:
            raise ValueError("Unknown dist: " + dist)

        if params["OPTIMIZE_CONSTANTS"]:
            # if we are training, then optimize the constants by
            # gradient descent and save the resulting phenotype
            # string as ind.phenotype_with_c0123 (eg x[0] +
            # c[0] * x[1]**c[1]) and values for constants as
            # ind.opt_consts (eg (0.5, 0.7). Later, when testing,
            # use the saved string and constants to evaluate.
            if dist == "training":
                return optimize_constants(x, y, ind)

            else:
                # this string has been created during training
                phen = ind.phenotype_consec_consts
                c = ind.opt_consts
                # phen will refer to x (ie test_in), and possibly to c
                yhat = eval(phen)
                assert np.isrealobj(yhat)

                # let's always call the error function with the
                # true values first, the estimate second
                return params["ERROR_METRIC"](y, yhat)

        else:
            # phenotype won't refer to C
            yhat = eval(ind.phenotype)
            assert np.isrealobj(yhat)

            # let's always call the error function with the true
            # values first, the estimate second
            return params["ERROR_METRIC"](y, yhat)
